
if { [info command dict] eq "" } {
    package require dict
}
package require tdom
package require msgcat
package require snit
package require math::constants
package require compass_utils

if { 0 } {
    package require xml2pdf
    package require tooltip
}

package provide compass_utils 1.1

proc isdouble { txt } {
    # this is to avoid an error in TCL 8.5a4 with 'na'
    if { [string index $txt 0] eq "n" } { return }
    return [string is double -strict $txt]
}

proc isdouble_or_void { txt } {
    # this is to avoid an error in TCL 8.5a4 with 'na'
    if { [string index $txt 0] eq "n" } { return }
    return [string is double $txt]
}

proc setAttributes { domNode xpath attList } {
    set inum 0
    foreach node [$domNode selectNodes $xpath] {
	foreach "name value" $attList {
	    $node setAttribute $name $value
	}
	incr inum
    }
    return $inum    
}

namespace eval formulae {
    variable nicer_name
    variable uvalues
    variable uvalues_moredata
    variable lognoter_db ""
}

################################################################################
#    General utilities
################################################################################

proc formulae::xml { args } {
    set map [list > "&gt;" < "&lt;" & "&amp;" \" "&quot;" ' "&apos;"]
    return [string map $map [join $args ""]]
}

proc formulae::subst_units { doc x } {

    regexp {\$([^\\]*)\\?\[(.*)\]} $x {} var unit_to
    regsub -all {\$} $unit_to {} unit_to
    set root [$doc documentElement]
    set node [$root selectNodes //param\[@n='$var'\]]
    set unit_from [$node @units]

    set degtorad [::math::constants::give_constant degtorad]
    set pi [::math::constants::give_constant pi]

    set units [list \
	    m 1.0 cm 1e-2 mm 1e-3 \
	    rad 1.0 deg $degtorad \
	    N 1.0 kN 1.0e3 \
	    N/m\u0b2 1.0 N/mm\u0b2 1.0e6 kN/m\u0b2 1.0e+3 \
	    W 1.0 MW 1.0e6 \
	    rad/s 1.0 rpm [expr {2*$pi/60.0}] \
	    ]

    if { ![dict exists $units $unit_from] } {
	error "error: unit $unit_from is not known"
    }
    if { ![dict exists $units $unit_to] } {
	error "error: unit $unit_to is not known"
    }
    set ex "(\$$var*[dict get $units $unit_from]/[dict get $units $unit_to])"
    return $ex
}

proc formulae::parse_formula { doc formula } {
    regsub -all {\m[a-zA-Z]\w*(?!\(|\w|[\s\w.]*"|\])} $formula {$&} formula
    set cmd "formulae::subst_units $doc \$x"
    set formula [regmap {\$[a-zA-Z]\w*?\\?\[.*?\]} $formula x $cmd]
    regsub -all {\^} $formula {**} formula
    regsub -all {/} $formula {*1.0/} formula
    return $formula
}

proc formulae::nicer_name { name } {
    variable nicer_name
    
    if { [info exists nicer_name($name)] } {
	return $nicer_name($name)
    } else { return $name }
}

proc formulae::nicer_formula { formula } {
    regsub -all {sqrt\(} $formula "\u221a(" formula
    set formula [regmap {\m[a-zA-Z]\w*(?!\(|\w|"|\])} $formula x {formulae::nicer_name $x }]
    regsub -all {\^(\w+)} $formula {<sup>\1</sup>} formula
    regsub -all {==} $formula {=} formula

    regsub -all {\*\*([\w$]+)} $formula {<sup>\1</sup>} formula
    regsub -all {\*\*\(([^)]+)\)} $formula {<sup>\1</sup>} formula
    regsub -all {\*} $formula {} formula
    regsub -all {\m(\d+)\.0\M} $formula {\1} formula
    regsub -all {\mpi\(\)} $formula "\u3c0" formula
    regsub -all {exp\((.*?)\)} $formula {e<sup>\1</sup>} formula
    if { [regexp {^\s*case\((.*)\)\s*$} $formula {} args] } {
	set formula [eval nicer_case_func [split $args ,]]
    }
    return $formula
}

proc formulae::nicer_units { units } {
    regsub {(.*)/} $units {(\1)/} units
    return $units
}

proc formulae::nicer_case_func { args } {
    set ret ""
    foreach "cond val" $args {
	if { $val eq "" } {
	    set val $cond
	    set cond "[_ default]"
	}
	append ret "$val ($cond)\n"
    }
    return $ret
}

proc formulae::nicer_formula_max { args } {
    set valList ""
    foreach i $args {
	lappend valList [nicer_formula $i]
    }
    return "max([join $valList ,])"
}

proc formulae::nicer_formula_min { args } {
    set valList ""
    foreach i $args {
	lappend valList [nicer_formula $i]
    }
    return "min([join $valList ,])"
}

proc formulae::max { doc args } {
    variable window_doc
    set val ""
    foreach i $args {
	set iv [formulae eval [list expr [parse_formula $window_doc $i]]]
	if { $val eq "" || $iv > $val } {
	    set val $iv
	}
    }
    return $val
}

proc formulae::min { doc args } {
    variable window_doc
    set val ""
    foreach i $args {
	set iv [formulae eval [list expr [parse_formula $window_doc $i]]]
	if { $val eq "" || $iv < $val } {
	    set val $iv
	}
    }
    return $val
}

proc formulae::e { arg } {
    variable window_doc
    return [formulae eval [list expr [parse_formula $window_doc $arg]]]
}

proc formulae::process_container { nnode containerNode } {
    variable nicer_name

    foreach node [$containerNode selectNodes param|condition] {
	if { [$node nodeName] eq "param" } {
	    set txt "[$node @pn [$node @n]]="
	} else { set txt "" }

	if { [string index [$node @value] 0] == "\[" } {
	    set value [string range [$node @value] 1 end-1]
	    set valueE [string map [list \[ \\\[] $value]
	    
	    if { [lsearch [list "formatted expression" "text expression"] \
		[$node @field_type ""]] != -1 } {
		set err [catch { formulae eval $valueE } v]
	    } else {
		set err [catch { eval $valueE } v]
	    }
	    if { $err } {
		set v "!ERROR ($v)"
	    }
	    if { [string match "e *" $value] } {
		set dosimple 1
		if { [$node nodeName] eq "condition" && [regexp {[<>]} $value] } {
		    #set f [string range $value 2 end]
		    set f [lindex $value 1]
		    regexp {(.*)([<>]=?)(.*)} $f {} v1 sign v2
		    set v1_e [string map [list \[ \\\[] $v1]
		    set v2_e [string map [list \[ \\\[] $v2]
		    set err 0
		    if { !$err } { set err [catch { eval "e $v1_e" } v1_res] }
		    if { !$err } { set err [catch { eval "e $v2_e" } v2_res] }
		    if { !$err } { set err [catch { nicer_formula $v1 } v1_f] }
		    if { !$err } { set err [catch { nicer_formula $v2 } v2_f] }
		    if { !$err && [isdouble $v1_res] && [isdouble_or_void $v2_res]} {
		        if { ![string is double -strict $v1_f] } {
		            append txt "$v1_f=[format %.4g $v1_res] "
		        } else {
		            append txt "$v1_f "
		        }
		        append txt "$sign "
		        if { ![string is double -strict $v2_f] } {
		            append txt "$v2_f=[format %.4g $v2_res]"
		        } else {
		            append txt "$v2_f"
		        }
		        set dosimple 0
		    }
		}
		if { $dosimple } {
		    #set f [string range $value 2 end]
		    set f [lindex $value 1]
		    set err [catch { nicer_formula $f } ret]
		    if { !$err } {
		        append txt $ret
		    } else {
		        set errorMessage $ret
		        append txt [_ "!ERROR"]
		    }
		}
	    }
	} else {
	    set err [catch { expr {[$node @value]*1.0} } v]
	    if { $err } { set v [$node @value] }
	}
	if { [$node nodeName] eq "param" } {
	    formulae eval [list set [$node @n] $v]
	    set nicer_name([$node @n]) [$node @pn [$node @n]]
	    if { [string index $txt end] ne "=" } { append txt "=" }
	    if { [string is double -strict $v] } {
		append txt "[format %.4g $v]"
	    } else { append txt $v }
	    if { [$node hasAttribute units] } {
		append txt " [nicer_units [$node @units]]"
	    }
	} else {
	    if { [$node hasAttribute units] } {
		append txt " [nicer_units [$node @units]]"
	    }
	    if { $v eq "!ERROR" } {
		# nothing
	    } elseif { $v } {
		if { [$node @true ""] ne "" } {
		    append txt "     [$node @true] (ok)"
		} else {
		    append txt "     [_ OK#C#Valid] (ok)"
		}
	    } else {
		if { [$node @false ""] ne "" } {
		    append txt "     [$node @false] (fail)"
		} else {
		    append txt "     [_ {Fail#C#No Valid}] (fail)"
		}
	    }
	}
	
	set row [$nnode appendChildTag row]
	set em  [$row appendChildTag entry emphasis]
	if { [string match "formatted*" [$node @field_type ""]] } {
	    set entry [$em parentNode]
	    $entry setAttribute cols_justify justify
	    if { [string trim $v] eq "" } { set v "<lognoter/>" }
	    $em appendXML $v
	    set n0 [lindex [$em childNodes] 0]
	    foreach n [$n0 childNodes] {
		$em appendChild $n
	    }
	    $n0 delete
	} else {
	    $em setAttribute role formula
	    $em appendChildText $txt
	}
	set entry [$row appendChildTag entry]
	$entry appendChildText [$node @help ""]
    }

}


proc formulae::create_images {} {
    # nothing
}

proc formulae::header_contextual_menu { wp x y } {

    if { $wp eq "." } { set wp "" }
    catch { destroy $wp._cmenu }
    set menu [menu $wp._cmenu -tearoff 0]

    $menu add command -label [_ "Edit"] -command [list formulae::edit_header $wp]
    $menu add command -label [_ "Create/edit property"] -command [list formulae::edit_properties \
	    $wp]
    $menu add separator
    $menu add command -label [_ "Delete all"] -command [list formulae::delete_full_formulae \
	    $wp]

    tk_popup $menu $x $y
}

proc formulae::entry_contextual_menu { wp field_name container_param x y } {
    
    if { $wp eq "." } { set wp "" }
    catch { destroy $wp._cmenu }
    set menu [menu $wp._cmenu -tearoff 0]
    
    $menu add command -label [_ "Edit"] -command [list formulae::edit_properties \
	    -field_name $field_name $wp]
    $menu add command -label [_ "Copy"] -command [list formulae::edit_properties \
	    -field_name $field_name -location [list after $field_name] -copy 1 $wp]
    $menu add separator
    if { $container_param eq "container" } {
	$menu add command -label [_ "Insert inside"] -command [list formulae::edit_properties \
		-location [list inside $field_name] $wp]
    }
    $menu add command -label [_ "Insert before"] -command [list formulae::edit_properties \
	    -location [list before $field_name] $wp]
    $menu add command -label [_ "Insert after"] -command [list formulae::edit_properties \
	    -location [list after $field_name] $wp]
    $menu add separator
    $menu add command -label [_ "Create drawing"] -command \
	[list formulae::addmodify_SVG_drawing $wp  $field_name]
    $menu add separator
    $menu add command -label [_ "Delete"] -command [list \
	    formulae::edit_properties_delete -field_name $field_name $wp]
    tk_popup $menu $x $y
}

proc formulae::add_drawing_contextual_menu { container_name menu } {
    
    $menu add separator
    $menu add command -label [_ "Edit drawing"] -command \
	[list formulae::addmodify_SVG_drawing $menu $container_name]
    $menu add separator
    $menu add command -label [_ "Delete"] -command \
	    [list formulae::delete_SVG_drawing $menu $container_name]
}

proc formulae::delete_SVG_drawing {  wp field_name } {
    variable window_doc

    set xp {
	//param[@n='N'] |
	//condition[@n='N'] |
	//container[@n='N']
    }
    set xp [string map [list N $field_name] $xp]
    set childNode [$window_doc selectNodes $xp]
    if { [$childNode nodeName] ne "container" } {
	set parentNode [$childNode parentNode]
    } else {
	set parentNode $childNode
    }

    set ns { svg http://www.w3.org/2000/svg }
    set svgNode [$parentNode selectNodes -namespaces $ns svg:svg]
    if { $svgNode ne "" } { $svgNode delete }
    create_windowD_do
}

proc formulae::addmodify_SVG_drawing { wp field_name } {
    variable window_doc

    set xp {
	//param[@n='N'] |
	//condition[@n='N'] |
	//container[@n='N']
    }
    set xp [string map [list N $field_name] $xp]
    set childNode [$window_doc selectNodes $xp]
    if { [$childNode nodeName] ne "container" } {
	set parentNode [$childNode parentNode]
    } else {
	set parentNode $childNode
    }

    set ns { svg http://www.w3.org/2000/svg }
    set svgNode [$parentNode selectNodes -namespaces $ns svg:svg]

    if { [interp exists ramdrawer] } { interp delete ramdrawer }
    interp create ramdrawer
    ramdrawer eval [list set argv [list -read_default_file 0 -save_callback \
		[list save_formulae $parentNode]]]
    if { $svgNode ne "" } {
	ramdrawer eval [list lappend argv -xmldata [list [$svgNode asXML] 1 *f*]]
    } else {
	ramdrawer eval [list lappend argv -xmldata [list "" 0 *f*]]
    }
    package require ramdrawer
    ramdrawer eval [list proc save_formulae { parentNode xml } {
	    master [list formulae::addmodify_SVG_drawing_do $parentNode $xml]
	}]
}

proc formulae::addmodify_SVG_drawing_do { parentNode xml } {

    dom parse $xml doc
    set svgNode [$doc documentElement]
    set ns { svg http://www.w3.org/2000/svg }

    set err [catch {
	    set svgNode_old [$parentNode selectNodes -namespaces $ns svg:svg]
	    if { $svgNode_old ne "" } { $svgNode_old delete }
	    $parentNode appendChild $svgNode
	} errstring]
    if { $err } {
	snit_messageBox -message [_ "Ramdrawer out of sync (%s)" $errstring]
    }
    create_windowD_do
}

proc formulae::_find_combo_with_var { w varname } {

    if { [winfo class $w] eq "TCombobox" && [$w cget -textvariable] eq $varname } {
	return $w
    }
    foreach i [winfo children $w] {
	set ret [_find_combo_with_var $i $varname]
	if { $ret ne "" } { return $ret }
    }
    return ""
}

proc formulae::_find_radiobuttons_with_var { w varname } {

    set rets ""
    if { [winfo class $w] eq "TRadiobutton" && [$w cget -variable] eq $varname } {
	lappend rets $w
    }
    foreach i [winfo children $w] {
	set rets_in [_find_radiobuttons_with_var $i $varname]
	if { [llength $rets_in] } { lappend rets {expand}$rets_in }
    }
    return $rets
}

proc formulae::_find_neighbor_labels_buttons { widget { class "" } } {

    set master [dict get [grid info $widget] -in]
    set row [dict get [grid info $widget] -row]
    set ret ""
    foreach child [grid slaves $master -row $row] {
	if { $child eq $widget } { continue }
	if { $class ne "" } {
	    if { [winfo class $child] eq $class } {
		lappend ret $child
	    }
	} elseif { [lsearch [list TLabel TButton TMenubutton] [winfo class $child]] != -1 } {
	    lappend ret $child
	}
    }
    return $ret
}

# update edit properties window
# args: ?-step prop_type|container? ?-field_name field_name? 
#       ?-original_field_name? ?-copy boolean? ?-location "before|after field_name"? w
proc formulae::edit_properties_update { args } {
    variable window_doc

    set steps [list "" prop_type container]
    set step_pos 0
    set field_name ""
    set copy_field 0
    set location ""
    while { [string match -* [lindex $args 0]] } {
	switch -- [lindex $args 0] {
	    -step {
		set args [lrange $args 1 end]
		set step_pos [lsearch -exact $steps [lindex $args 0]]
	    }
	    -field_name {
		set args [lrange $args 1 end]
		set field_name [lindex $args 0]
	    }
	    -original_field_name {
		set w [lindex $args end]
		set field_name [$w give_uservar_value original_field_name]
		if { $location eq "" } {
		    set location [$w give_uservar_value original_location]
		}
	    }
	    -location {
		set args [lrange $args 1 end]
		set location [lindex $args 0]
	    }
	    -copy {
		set args [lrange $args 1 end]
		set copy_field [lindex $args 0]
	    }
	    default { error "error in formulae::edit_properties_update" }
	}
	set args [lrange $args 1 end]
    }
    foreach "w" $args break
    
    foreach i [list param_fields goto_fields fields sfields] {
	set $i ""
    }
    set containers ""
    foreach domNode [$window_doc selectNodes //param|//condition|//container] {
	if { [$domNode hasAttribute n] } {
	    if { [$domNode hasAttribute n] } {
		if { [$domNode nodeName] eq "container" && [llength $goto_fields] } {
		    lappend goto_fields ---
		}
		lappend goto_fields [$domNode @n]
	    }
	    if { [$domNode hasAttribute n] } {
		lappend fields [$domNode @n]
	    }
	    if { [$domNode nodeName] eq "param" && [$domNode hasAttribute n] } {
		lappend param_fields [$domNode @n]
	    }
	    if { [$domNode hasAttribute pns] } {
		lappend sfields [$domNode @pns [$domNode @n]]
	    }
	    if { [$domNode nodeName] eq "container" } {
		lappend containers [$domNode @n]
	    }
	}
    }

    if { !$copy_field } {
	$w set_uservar_value original_field_name $field_name
    } else {
	$w set_uservar_value original_field_name ""
	$w set_uservar_value copyfrom_field_name $field_name
    }
    $w set_uservar_value original_location $location

    set xp "//*\[@n='$field_name' and not(ancestor::setname)\]"
    set node [$window_doc selectNodes $xp]

################################################################################
#    property type
################################################################################

    if { $step_pos == 0 || (![llength $containers] && \
	[$w give_uservar_value prop_type] ne "container") } {
	if { ![llength $containers] } {
	    $w set_uservar_value prop_type container
	} elseif { $field_name eq "" } {
	    $w set_uservar_value prop_type param
	} else {
	    $w set_uservar_value prop_type [$node nodeName]
	}
    }
    foreach r [_find_radiobuttons_with_var $w [$w give_uservar prop_type]] {
	if { $field_name ne "" } {
	    $r state disabled
	} elseif { [llength $containers] || [$r cget -value] eq "container" } {
	    $r state !disabled
	} else {
	    $r state disabled
	}
    }

################################################################################
#    navigate
################################################################################

    set b [$w give_uservar_value previous_button]
    $b configure -command [list formulae::edit_properties_navigate $w $field_name prev]

    set b [$w give_uservar_value next_button]
    $b configure -command [list formulae::edit_properties_navigate $w $field_name next]

    set goto_menu [$w give_uservar_value goto_menu]
    $goto_menu delete 0 end
    foreach i $goto_fields {
	if { $i eq "---" } {
	    $goto_menu add separator
	} else {
	    $goto_menu add command -label $i -command \
		[list formulae::edit_properties_navigate $w $i this]
	}
    }

################################################################################
#    location
################################################################################

    if { [llength $location] && [lindex $location 1] ne "" } {
	set xp {
	    //param[@n='N'] |
	    //condition[@n='N'] |
	    //container[@n='N']
	}
	set xp1 [string map [list N [lindex $location 1]] $xp]
	set location_node [$window_doc selectNodes $xp1]
	set container [$location_node selectNodes {string(../@n)}]
	if { [$w give_uservar_value prop_type] eq "container" } {
	    $w set_uservar_value container ""
	} elseif { $step_pos == 2 } {
	    #nothing
	} elseif { $container ne "" } {
	    $w set_uservar_value container $container
	} elseif { [lindex $location 0] ne "inside" } {
	    $w set_uservar_value container ""
	    $w set_uservar_value prop_type container
	} else {
	    $w set_uservar_value container [$location_node @n]
#[lindex $containers 0]
	}
    } elseif { $step_pos <= 1 } {
	if { $field_name eq "" } {
	    if { [$w give_uservar_value prop_type] eq "container" } {
		$w set_uservar_value container ""
	    } else {
		$w set_uservar_value container [lindex $containers 0]
	    }
	} else {
	    $w set_uservar_value container [$node selectNodes {string(../@n)}]
	}
    }

    set c [_find_combo_with_var $w [$w give_uservar container]]
    if { [$w give_uservar_value prop_type] eq "container" } {
	$c configure -values ""
	$c state disabled
    } else {
	$c state !disabled
	$c configure -values $containers
    }

    set container [$w give_uservar_value container]
    set containerNode [$window_doc selectNodes //container\[@n='$container'\]]
    if { $containerNode eq "" } {
	set containerNode [$window_doc selectNodes formulae]
    }
    set nodes [$containerNode selectNodes container|param|condition]
    set ipos [lsearch -exact $nodes $node]

    if { [llength $location] && [lindex $location 1] ne ""  } {
	if { [lsearch -exact $nodes $location_node] != -1 } {
	    switch [lindex $location 0] {
		before {
		    $w set_uservar_value before_after [_ Before]
		}
		after {
		    $w set_uservar_value before_after [_ After]
		}
	    }
	    set nei_node $location_node
	} else {
	    $w set_uservar_value before_after [_ After]
	    set nei_node [lindex $nodes end]
	}
    } elseif { [$w give_uservar_value before_after] eq [_ After] && $ipos > 0 } {
	set nei_node [lindex $nodes [expr {$ipos-1}]]
    } elseif { $ipos>= 0 && $ipos < [llength $nodes]-1 } {
	$w set_uservar_value before_after [_ Before]
	set nei_node [lindex $nodes [expr {$ipos+1}]]
    } elseif { [llength $nodes] && $ipos == -1 } {
	$w set_uservar_value before_after [_ After]
	set nei_node [lindex $nodes end]
    } else { set nei_node "" }

    if { $nei_node ne "" } {
	$w set_uservar_value before_after_name [$nei_node @n ""]
    } else {
	$w set_uservar_value before_after_name ""
    }

    set c [_find_combo_with_var $w [$w give_uservar before_after_name]]
    set values ""
    foreach i $nodes {
	if { [$i hasAttribute n] } {
	    lappend values [$i @n]
	}
    }
    $c configure -values $values

    set mb [formulae::_find_neighbor_labels_buttons $c TMenubutton]
    if { [$w give_uservar_value prop_type] eq "container" } {
	$mb state !disabled
    } else {
	$w set_uservar_value same_page 0
	$mb state disabled
    }
    if { [lindex $steps $step_pos] eq "container" } {
	# a shortcut to avoid changing user modified values
	return
    }

    if { !$copy_field && $field_name ne "" && [$node @location ""] eq "same_page" } {
	$w set_uservar_value same_page 1
    } else {
	$w set_uservar_value same_page 0
    }

################################################################################
#    Name
################################################################################

    if { !$copy_field && $field_name ne "" } {
	$w set_uservar_value name [$node @n ""]
	set state disabled
    } else {
	$w set_uservar_value name ""
	set state !disabled
    }
    set c [_find_combo_with_var $w [$w give_uservar name]]
    $c configure -values $fields
    $c state $state

################################################################################
#    short name
################################################################################

    set c [_find_combo_with_var $w [$w give_uservar short_name]]
    if { [$w give_uservar_value prop_type] eq "container" } {
	$c state !disabled
	foreach label [_find_neighbor_labels_buttons $c] {
	    $label state !disabled
	}
	if { $field_name ne "" } {
	    $w set_uservar_value short_name [$node @pns [$node @n]]
	}
	$c configure -values $sfields
    } else {
	$c state disabled
	foreach label [_find_neighbor_labels_buttons $c] {
	    $label state disabled
	}
    }

################################################################################
#    print name
################################################################################

    set c [$w give_uservar_value print_name_widget]
    $c ClearText
    if { [$w give_uservar_value prop_type] eq "condition" } {
	$c configure -state disabled -bg grey90
	foreach label [_find_neighbor_labels_buttons $c] {
	    $label state disabled
	}
    } else {
	$c configure -state normal -bg white
	if { $field_name ne "" } {
	    $c ApplyXMLdataAsText <lognoter>[$node @pn [$node @n]]</lognoter>
	}
	foreach label [_find_neighbor_labels_buttons $c] {
	    $label state !disabled
	}
    }

################################################################################
#    field type & expression
################################################################################

#     set field_types [list numeric text "long text" expression "text expression" \
#             "options editable" "options non-editable" formatted \
#             "formatted expression" button date file]

    set value ""
    if { [$w give_uservar_value prop_type] ne "container" } {
	if { $field_name ne "" } { set value [$node @value ""] }
    } else {
	if { $field_name ne "" } { set value [$node @condition ""] }
	$w set_uservar_value value_label [_ Condition:]
    }

    set c [_find_combo_with_var $w [$w give_uservar field_type]]

    set field_type [$w give_uservar_value field_type numeric]

    if { [$w give_uservar_value prop_type] eq "param" } {
	$c state !disabled
	foreach label [_find_neighbor_labels_buttons $c] {
	    $label state !disabled
	}
	if { $field_name ne "" } {
	    if { [$node @field_type ""] eq "" } {
		if { [$node @values ""] ne "" } {
		    if { [$node @editable 0] } {
		        set field_type "options editable"
		    } else {
		        set field_type "options non-editable"
		    }
		} elseif { [regexp {\[e\s+(.*)\]} $value] } {
		    set field_type expression
		} elseif { [string is double -strict $value] } {
		    set field_type numeric
		} else {
		    set field_type text
		}
	    } else {
		set field_type [$node @field_type]
	    }
	}
    } else {
	$c state disabled
	foreach label [_find_neighbor_labels_buttons $c] {
	    $label state disabled
	}
	set field_type expression
    }

    switch $field_type {
	"expression" {
	    set cmd_in [string trim $value {[]}]
	    set err [catch { lindex $cmd_in 0 } c]
	    if { !$err && $c eq "e" } {
		set value [lindex $cmd_in 1]
	    } else {
		regexp {\[e\s+(.*)\]} $value {} value
		set value [string trim $value "{}"]
	    }
	}
	"text expression" {
	    regexp {\[(.*)\]} $value {} value
	}        
	"formatted expression" {
	    regexp {\[(.*)\]} $value {} value
	}
	"options editable" - "options non-editable" {
	    $w set_uservar_value valuesList [$node @values]
	}
	file {
	    set value ""
	}
    }
    $w set_uservar_value field_type $field_type
    $w set_uservar_value value $value

    set value_widget [$w give_uservar_value value_widget]
    set value_fields_widget [$w give_uservar_value value_fields_widget]
    $value_fields_widget delete 0 end
    foreach i $param_fields {
	$value_fields_widget add command -label $i -command \
	    [list formulae::_insert_in_entry $value_widget $i]
    }

################################################################################
#    units
################################################################################

    if { $field_name ne "" } {
	$w set_uservar_value units [$node @units ""]
    } else { $w set_uservar_value units "" }
    set c [_find_combo_with_var $w [$w give_uservar units]]
    if { [$w give_uservar_value prop_type] ne "container" } {
	$c state !disabled
	foreach label [_find_neighbor_labels_buttons $c] {
	    $label state !disabled
	}
    } else {
	$c state disabled
	foreach label [_find_neighbor_labels_buttons $c] {
	    $label state disabled
	}
    }

################################################################################
#    true & false
################################################################################

    if { $field_name ne "" } {
	$w set_uservar_value true [$node @true ""]
	$w set_uservar_value false [$node @false ""]
    } else {
	$w set_uservar_value true ""
	$w set_uservar_value false ""
    }
    set c1 [_find_combo_with_var $w [$w give_uservar true]]
    set c2 [_find_combo_with_var $w [$w give_uservar false]]
    if { [$w give_uservar_value prop_type] eq "condition" } {
	$c1 state !disabled
	$c2 state !disabled
	foreach label [_find_neighbor_labels_buttons $c1] {
	    $label state !disabled
	}
	foreach label [_find_neighbor_labels_buttons $c2] {
	    $label state !disabled
	}
    } else {
	$c1 state disabled
	$c2 state disabled
	foreach label [_find_neighbor_labels_buttons $c1] {
	    $label state disabled
	}
	foreach label [_find_neighbor_labels_buttons $c2] {
	    $label state disabled
	}
    }

################################################################################
#    help
################################################################################

    set widget [$w give_uservar_value help_widget]
    $widget delete 1.0 end
    if { $field_name ne "" } {
	$widget insert end [$node @help ""]
    }
}

# accept values in edit properties
# args: ?-refresh boolean? ?-delete_original? ?-copy_children boolean? w
proc formulae::edit_properties_accept { args } {
    variable window_doc

    set delete_original_field_name 1
    set refresh 1
    set copy_children 0
    while { [string match -* [lindex $args 0]] } {
	switch -- [lindex $args 0] {
	    -refresh {
		set args [lrange $args 1 end]
		set refresh [lindex $args 0]
	    }
	    -copy_children {
		# this is not still implemented
		set args [lrange $args 1 end]
		set copy_children [lindex $args 0]
	    }
	    -delete_original { set delete_original_field_name 0 }
	    default { error "error in formulae::edit_properties_accept" }
	}
	set args [lrange $args 1 end]
    }
    foreach "w" $args break

    set prop_type [$w give_uservar_value prop_type]
    set newnode [[$window_doc ownerDocument] createElement $prop_type]

    set original_field_name [$w give_uservar_value original_field_name]

################################################################################
#    location
################################################################################

    if { $prop_type eq "container" && [$w give_uservar_value same_page] } {
	$newnode setAttribute location same_page
    }

################################################################################
#    name
################################################################################

    set name [string trim [$w give_uservar_value name]]
    if { $name eq "" && $prop_type eq "condition" } {
	set name [string trim [$w give_uservar_value value]]
    }
    if { $name eq "" } {
	$newnode delete
	error [_ "error: name cannot be void"]
    }
    if { [regexp {__} $name] } {
	$newnode delete
	error [_ "error: name cannot contain 2 underlines (__)"]
    }

    if { ![regexp {^[a-zA-Z]\w*$} $name] } {
	error [_ "error: name must begin with a letter and contain letters, numbers or _"]
    }

    $newnode setAttribute n $name

    if { $prop_type eq "container" } {
	$newnode setAttribute pns [string trim [$w give_uservar_value short_name]]
    }

################################################################################
#    print name
################################################################################

    set c [$w give_uservar_value print_name_widget]
    if { $prop_type ne "condition" } {
	dom parse <lognoter>[$c exportXML]</lognoter> doc_pn
	$doc_pn documentElement root_pn
	set paraNode [$root_pn selectNodes {para[1]}]
	set pn ""
	foreach node [$paraNode childNodes] {
	    append pn [$node asXML]
	}
	set pn [string trim $pn]
	if { $pn ne "" } {
	    $newnode setAttribute pn $pn
	} elseif { [$newnode hasAttribute pn] } {
	    $newnode removeAttribute pn
	}
    }

################################################################################
#    expression & condition
################################################################################

#     set field_types [list numeric text "long text" expression "text expression" \
#             "options editable" "options non-editable" formatted \
#             "formatted expression" button date file]

    foreach att [list field_type values editable command] {
	if { [$newnode hasAttribute $att] } {
	    $newnode removeAttribute $att
	}
    }
    if { $prop_type eq "param" } {
	$newnode setAttribute field_type [$w give_uservar_value field_type]
    } else { set field_type expression }

    set value [string trim [$w give_uservar_value value]]

    switch [$w give_uservar_value field_type] {
	"expression" {
	    if { $value eq "" } {
		if { $prop_type ne "container" } {
		    $newnode delete
		    error [_ "error: expression cannot be void"]
		}
	    } else {
		set value "\[[list e $value]\]"
	    }
	}
	"text expression" - "formatted expression" {
	    if { $value eq "" } {
		$newnode delete
		error [_ "error: expression cannot be void"]
	    } 
	    set value "\[$value\]"
	}
	"options editable" - "options non-editable" {
	    $newnode setAttribute values [string trim [$w give_uservar_value \
		valuesList]]
	    if { [$w give_uservar_value field_type] eq "options editable" } {
		$newnode setAttribute editable 1
	    } else {
		if { [lsearch [split [$newnode @values] ,] $value] == -1 } {
		    $newnode delete
		    error [_ "error: default value for option must be one element of the list"]
		}
		$newnode setAttribute editable 0
	    }
	}
	"numeric" {
	    if { $value eq "" } {
		set value 0.0
	    } elseif { ![string is double -strict $value] } {
		$newnode delete
		error [_ "error: value must be a number"]
	    }
	}
	"button" {
	    $newnode setAttribute command $value
	}
	file {
	    set value ""
	}
    }
    if { $prop_type ne "container" } {
	$newnode setAttribute value $value
    } else {
	$newnode setAttribute condition $value
    }


################################################################################
#    units & true/false & help
################################################################################

    if { $prop_type ne "container" } {
	$newnode setAttribute units [$w give_uservar_value units]
    }
    if { $prop_type eq "condition" } {
	$newnode setAttribute true [$w give_uservar_value true]
	$newnode setAttribute false [$w give_uservar_value false]
    }
    set widget [$w give_uservar_value help_widget]
    $newnode setAttribute help [string trim [$widget get 1.0 end-1c]]

################################################################################
#    deleting old nodes & inserting new one
################################################################################

    set xp {
	//param[@n='N'] |
	//condition[@n='N'] |
	//container[@n='N']
    }
    set xp1 [string map [list N $original_field_name] $xp]
    if { $delete_original_field_name } {
	foreach domNode [$window_doc selectNodes $xp1] {
	    foreach childNode [$domNode childNodes] {
		$newnode appendChild [$domNode removeChild $childNode]
	    }
	    $domNode delete
	}
    }
    set xp2 [string map [list N $name] $xp]
    foreach domNode [$window_doc selectNodes $xp2] {
	foreach childNode [$domNode childNodes] {
	    $newnode appendChild [$domNode removeChild $childNode]
	}
	$domNode delete
    }
    set container [$w give_uservar_value container]
    set containerNode [$window_doc selectNodes //container\[@n='$container'\]]
    if { $containerNode eq "" } {
	set containerNode [$window_doc selectNodes formulae]
    }

    set nei_node_name [$w give_uservar_value before_after_name]
    set xp2 [string map [list N $nei_node_name] $xp]
    set nei_node [$window_doc selectNodes $xp2]
    if { $nei_node ne "" && [$w give_uservar_value before_after] eq [_ After] } {
	set nodes [$containerNode childNodes]
	set ipos [lsearch -exact $nodes $nei_node]
	if { $ipos == -1 || $ipos == [llength $nodes]-1 } {
	    set nei_node ""
	} else {
	    set nei_node [lindex $nodes [incr ipos]]
	}
    }
    if { $nei_node eq "" } {
	$containerNode appendChild $newnode
    } else {
	$containerNode insertBefore $newnode $nei_node
    }

    create_update_database_table

    if { $refresh } {
	create_windowD_do
    } else {
	$w set_uservar_value refresh_pending 1
    }
}

# args: ?-refresh boolean? ?-field_name field_name? w
proc formulae::edit_properties_delete { args } {
    variable window_doc

    set refresh 1
    set field_name ""
    while { [string match -* [lindex $args 0]] } {
	switch -- [lindex $args 0] {
	    -field_name {
		set args [lrange $args 1 end]
		set field_name [lindex $args 0]
	    }
	    -refresh {
		set args [lrange $args 1 end]
		set refresh [lindex $args 0]
	    }
	    default { error "error in formulae::edit_properties_delete" }
	}
	set args [lrange $args 1 end]
    }
    foreach "w" $args break

    if { $field_name eq "" } {
	set field_name [string trim [$w give_uservar_value name]]
    }
    if { $field_name eq "" } {
	error [_ "error: name cannot be void"]
    }
    set xp {
	//param[@n='N'] |
	//condition[@n='N'] |
	//container[@n='N']
    }
    set xp1 [string map [list N $field_name] $xp]
    set idx 0
    foreach domNode [$window_doc selectNodes $xp1] {
	$domNode delete
	incr idx
    }
    if { !$idx } {
	error [_ "error: name is not correct"]
    }
    if { $refresh } {
	create_windowD_do
    } else {
	$w set_uservar_value refresh_pending 1
    }
}

# edit header
# args: ?-update_callback update_callback? wp
proc formulae::edit_header { args } {
    variable window_doc

    package require tile

    set update_callback ""
    while { [string match -* [lindex $args 0]] } {
	switch -- [lindex $args 0] {
	    -update_callback {
		set args [lrange $args 1 end]
		set update_callback [lindex $args 0]
	    }
	    default { error "error in formulae::edit_header" }
	}
	set args [lrange $args 1 end]
    }
    foreach "wp" $args break


    destroy $wp._props
    set w [dialogwin_snit $wp._props -title [_ "Change header"] \
	    -callback [list formulae::edit_header_do $update_callback] -grab 0 \
	    -morebuttons [list [_ Delete]]]

    set f [$w giveframe]

    ttk::label $f.l1 -text [_ Title:]
    ttk::entry $f.e1 -textvariable [$w give_uservar title] -width 40

    ttk::label $f.l2 -text [_ Description:]
    text $f.t1 -wrap word -width 60 -height 4
    $w set_uservar_value desc_widget $f.t1

    bind $f.t1 <Return> "[bind Text <Return>];break"
    bind $f.t1 <Tab> "[bind all <Tab>];break"
    bind $f.t1 <<PrevWindow>> "[bind all <<PrevWindow>>];break"

    set tabstyles [dict create "" "" arrows [_ Arrows] arrows_small \
	    [_ "Small arrows"] notebook [_ Notebook]]
    $w set_uservar_value tabstyles_inv [dict_inverse $tabstyles]

    ttk::label $f.l3 -text [_ "Pages style:"]
    ttk::combobox $f.cb0 -textvariable [$w give_uservar tabstyle] \
	-values [dict values $tabstyles] -state readonly

    ttk::checkbutton $f.cb1 -text [_ "Has loadcase name"] -variable \
	[$w give_uservar hassetname]

    ttk::frame $f.f1 -borderwidth 0
    ttk::checkbutton $f.f1.cb1 -text [_ "Has database"] -variable \
	[$w give_uservar hasdatabase]
    ttk::checkbutton $f.f1.l1 -text [_ "Database name"]: -variable \
	[$w give_uservar hasdatabasename]
    ttk::entry $f.f1.e1 -textvariable [$w give_uservar databasename]

    grid $f.f1.cb1 $f.f1.l1 $f.f1.e1 -sticky w
    grid configure $f.f1.l1 -padx "20 0"

    set d [dict create 0 "$f.f1.cb1 $f.f1.l1" 1 ""]
    $w enable_disable_on_key hassetname $d
    set d [dict create 1 "hasdatabase 0"]
    $w change_key_on_key hassetname $d

    set d [dict create 0 "$f.cb1" 1 ""]
    $w enable_disable_on_key hasdatabase $d
    set d [dict create 1 "hassetname 0"]
    $w change_key_on_key hasdatabase $d

    set d [dict create 0 "" 1 "$f.f1.e1"]
    $w enable_disable_on_key hasdatabasename $d

    ttk::checkbutton $f.cb2 -text [_ "Usable in view mode"] -variable \
	[$w give_uservar usableviewmode]

    ttk::checkbutton $f.cb3 -text [_ "Only form in view mode"] -variable \
	[$w give_uservar only_form_view_mode]

    tooltip $f.cb3 [_ "If this option is selected, the notes part of the page is not visualized in view mode"]

    if { [info exists window_doc] } {
	set title [string trim [$window_doc selectNodes string(formulae/title)]]
	$w set_uservar_value title $title
	set desc [string trim [$window_doc selectNodes string(formulae/description)]]
	$f.t1 insert end $desc
	set setnameNode [$window_doc selectNodes formulae/setname]
	if { $setnameNode eq "" } {
	    $w set_uservar_value hassetname 0
	} else {
	    $w set_uservar_value hassetname 1
	}
	set formulaeNode [$window_doc selectNodes formulae]
	$w set_uservar_value usableviewmode [$formulaeNode @usableviewmode 0]
	$w set_uservar_value only_form_view_mode [$formulaeNode @only_form_view_mode 0]

	$w set_uservar_value tabstyle [dict get $tabstyles \
		[$formulaeNode @tabstyle ""]]
	if { [$formulaeNode @default_database 0] == 1 } {
	    $w set_uservar_value hasdatabase 1
	    $w set_uservar_value hasdatabasename 0
	    $w set_uservar_value databasename ""
	} elseif { [$formulaeNode @database ""] eq "" } {
	    $w set_uservar_value hasdatabase 0
	    $w set_uservar_value hasdatabasename 0
	    $w set_uservar_value databasename ""
	} else {
	    $w set_uservar_value hasdatabase 1
	    $w set_uservar_value hasdatabasename 1
	    $w set_uservar_value databasename [$formulaeNode @database]
	}
    } else {
	$w set_uservar_value title ""
	$w set_uservar_value hassetname 0
	$w set_uservar_value usableviewmode 0
	$w set_uservar_value only_form_view_mode 0
	$w set_uservar_value tabstyle ""
	$w set_uservar_value hasdatabase 0
	$w set_uservar_value hasdatabasename 0
	$w set_uservar_value databasename ""
    }

    grid $f.l1 $f.e1 -sticky w -pady 2
    grid $f.l2 - -sticky w -padx "40 0" -pady 2
    grid $f.t1 - -sticky ew -padx "40 2"
    grid $f.l3 $f.cb0 -sticky w -pady 2
    grid $f.cb1 - -sticky w -pady 2
    grid $f.f1  - -sticky w -pady 2
    grid $f.cb2 - -sticky w -pady 2
    grid $f.cb3 - -sticky w -pady 2
    grid configure $f.e1 -sticky ew
    grid columnconfigure $f 1 -weight 1

    tk::TabToWindow $f.e1
    bind $w <Return> [list $w invokeok]
    
    $w createwindow

}

proc formulae::edit_header_do { update_callback w } {
    variable window_doc
    variable lognoter_db
    variable database_default_table

    set action [$w giveaction]

    switch -- $action {
	1 {
	    if { ![info exists window_doc] } {
		set window_doc [dom createDocument formulae]
	    }
	    set formulaeNode [$window_doc selectNodes formulae]

	    if { [$w give_uservar_value hasdatabase] } {
		set needs_update 0
		if { ![$w give_uservar_value hasdatabasename] } {
		    set databasename $database_default_table
		    if { [$formulaeNode @default_database 0] == 0 } {
		        set needs_update 1
		    }
		} else {
		    set databasename [string trim [$w give_uservar_value databasename]]
		    if { $databasename eq "" } {
		        snit_messageBox -icon error -message \
		            [_ "It is necessary to enter a database name"]
		        return
		    }
		    if { ![regexp {^[\w-]+$} $databasename] } {
		        snit_messageBox -icon error -message \
		            [_ "Database name must only contain letters, numbers -_"]
		        return
		    }
		    if { [$formulaeNode @default_database 0] == 1 } {
		        set needs_update 1
		    }
		    if { [$formulaeNode @database ""] ne $databasename } {
		        set needs_update 1
		    }
		}
		if { $needs_update } {
		    set delete_db ""
		    if { [$formulaeNode @database ""] ne "" && \
		        ![string equal -nocase [$formulaeNode @database] \
		                $databasename] } {
		        set ret [snit_messageBox -type yesnocancel -message \
		                [_ "Do you want to delete database '%s'?" \
		                    [$formulaeNode @database]]]
		        if { $ret eq "cancel" } { return }
		        if { $ret eq "yes" } {
		            set delete_db [$formulaeNode @database]
		        }
		    }                    
		    if { [$w give_uservar_value hasdatabasename] } {
		        $formulaeNode setAttribute default_database 0
		    } else {
		        $formulaeNode setAttribute default_database 1
		    }
		    set old_database [$formulaeNode @database ""]
		    $formulaeNode setAttribute database $databasename
		    create_update_database_table -old_database $old_database

		    if { $delete_db ne "" } {
		        $lognoter_db sql exec "drop table \"$delete_db\""
		    }
		}

	    } elseif { [$formulaeNode @database ""] ne "" } {
		set ret [snit_messageBox -type yesnocancel -message \
		        [_ "Do you want to delete database '%s'?" \
		            [$formulaeNode @database]]]
		if { $ret eq "cancel" } { return }
		if { $ret eq "yes" } {
		    $lognoter_db sql exec "drop table \"[$formulaeNode @database]\""
		}
		catch { $formulaeNode removeAttribute default_database }
		$formulaeNode removeAttribute database
	    }
	    set title [string trim [$w give_uservar_value title]]
	    if { $title eq "" } {
		snit_messageBox -icon error -message \
		    [_ "It is necessary to enter a title"]
		return
	    }
	    set titleNode [$formulaeNode selectNodes title]
	    if { $titleNode ne "" } {
		foreach i [$titleNode childNodes] { $i delete }
	    } else {
		set titleNode [$window_doc createElement title]
		set firstChild [$formulaeNode selectNodes {./*[1]}]
		if { $firstChild ne "" } {
		    $formulaeNode insertBefore $titleNode $firstChild
		} else {
		    $formulaeNode appendChild $titleNode
		}
	    }
	    $titleNode appendChildText $title

	    $formulaeNode setAttribute usableviewmode [$w give_uservar_value \
		    usableviewmode]
	    $formulaeNode setAttribute only_form_view_mode [$w give_uservar_value \
		    only_form_view_mode]

	    set tabstyles_inv [$w give_uservar_value tabstyles_inv]
	    $formulaeNode setAttribute tabstyle [dict get $tabstyles_inv \
		    [$w give_uservar_value tabstyle]]

	    set desc_widget [$w give_uservar_value desc_widget]
	    set description [string trim [$desc_widget get 1.0 end-1c]]
	    set descNode [$formulaeNode selectNodes description]
	    if { $descNode ne "" } {
		foreach i [$descNode childNodes] { $i delete }
	    } else {
		set descNode [$window_doc createElement description]
		set secondChild [$formulaeNode selectNodes {./*[2]}]
		if { $secondChild ne "" } {
		    $formulaeNode insertBefore $descNode $secondChild
		} else {
		    $formulaeNode appendChild $descNode
		}
	    }
	    $descNode appendChildText $description

	    set setnameNode [$formulaeNode selectNodes setname]
	    if { [$w give_uservar_value hassetname] && $setnameNode eq "" } {
		set setnameNode [$window_doc createElement setname]
		set thirdChild [$formulaeNode selectNodes {./*[3]}]
		if { $thirdChild ne "" } {
		    $formulaeNode insertBefore $setnameNode $thirdChild
		} else {
		    $formulaeNode appendChild $setnameNode
		}
	    } elseif { ![$w give_uservar_value hassetname] && $setnameNode ne "" } {
		$setnameNode delete
	    }

	    if { $update_callback ne "" } {
		uplevel #0 $update_callback
	    } else {
		create_windowD_do
	    }
	}
	2 {
	    set fail [delete_full_formulae $w]
	    if { $fail } { return }
	}
    }
    destroy $w
}

proc formulae::give_create_statement { formulaeNode lognoter_db } {

    set names ""
    set types ""
    set defaults ""
    foreach paramNode [$formulaeNode selectNodes //param] {
	if { [$paramNode @in_database 1] == 0 } { continue }
	set field_type [$paramNode @field_type ""]
	if { $field_type eq "button" } { continue }

	lappend names [$paramNode @n]

	if { $field_type eq "file" } {
	    lappend names [$paramNode @n]__contents [$paramNode @n]__cdate \
		[$paramNode @n]__mdate [$paramNode @n]__type [$paramNode @n]__size \
		[$paramNode @n]__path
	    lappend types text longblob datetime datetime text numeric text
	    lappend defaults "" "" "" "" "" ""
	} elseif { [lsearch "numeric expression" $field_type] != -1 || \
	    [$paramNode @units ""] ne "" || [string is double -strict [$paramNode @value]] ||
	    [regexp {^\[e} [$paramNode @value]] } {
	    lappend types numeric
	} else {
	    lappend types text
	}
	if { [regexp {^\[} [$paramNode @value]] } {
	    lappend defaults 0.0
	} else {
	    lappend defaults ""
	}
    }
    set cmdList ""
    foreach name $names type $types def $defaults {
	if { [string is double -strict def] } {
	    lappend cmdList "\"$name\" $type default $def"
	} else {
	    lappend cmdList "\"$name\" $type"
	}
    }
    if { [$lognoter_db cget -dbtype] eq "mysql" } {
	set id_type "integer unsigned auto_increment primary key"
    } else {
	set id_type "integer primary key"
    }
    set cmd "create table \"[$formulaeNode @database]\"(
	id $id_type,creation_date datetime,
	modification_date datetime"
    if { [llength $cmdList] } {
	append cmd ",[join $cmdList ,]"
    }
    append cmd ")"
    if { [$lognoter_db cget -dbtype] eq "mysql" } {
	append cmd " type = InnoDB"
    }
    return [list $names $cmd]
}

proc formulae::create_update_database_table { args } {
    variable window_doc
    variable lognoter_db

    set optional {
	{ -old_database db "" }
    }
    set compulsory ""
    parse_args $optional $compulsory $args

    set formulaeNode [$window_doc selectNodes formulae]
    if { [$formulaeNode @database ""] eq "" } { return }

    foreach "names cmd" [give_create_statement $formulaeNode $lognoter_db] break

    if { $old_database eq "" } {
	set old_database [$formulaeNode @database]
    }
    set table_to_delete ""
    if { $old_database ne "" && [$lognoter_db sql table_exists $old_database] } {
	set cols [$lognoter_db sql column_names $old_database]
	set old_names [lrange $cols 3 end]
	if { $names eq $old_names && $old_database eq [$formulaeNode @database] } {
	    return
	}
	if { $old_database eq [$formulaeNode @database] } {
	    catch { $lognoter_db sql exec "drop table \"${old_database}_old\"" }
	    $lognoter_db sql exec "alter table \"$old_database\"
		rename to \"${old_database}_old\""
	    set old_database [$formulaeNode @database]_old
	    set table_to_delete $old_database
	}
    } elseif { $old_database ne "" } { set old_database "" }

    catch { $lognoter_db sql exec "drop table \"[$formulaeNode @database]\"" }
    $lognoter_db sql exec $cmd

    if { $old_database ne "" } {
	if { [llength $names] == [llength $old_names] } {
	    $lognoter_db sql exec "insert into \"[$formulaeNode @database]\" select
		* from $old_database"
	} else {
	    set common_fields [list_intersection $names $old_names]
	    if { [llength $common_fields] } {
		$lognoter_db sql exec "insert into \"[$formulaeNode @database]\"
		    (\"[join $common_fields \",\"]\") select 
		    \"[join $common_fields \",\"]\" from
		    \"$old_database\""
	    }
	}
	if { $table_to_delete ne "" } {
	    $lognoter_db sql exec "drop table \"$table_to_delete\""
	}
    }
}

proc formulae::delete_full_formulae { w } {
    variable window_doc
    variable lognoter_db

    if { ![info exists window_doc] } {
	snit_messageBox -message [_ "Nothing to delete"] -parent $w
	return -1
    }
    set text [_ "Are you sure to delete ALL forms information in this page?"]
    set retval [snit_messageBox -default cancel -icon question -message $text \
	    -type okcancel -parent $w]
    if { $retval == "cancel" } {
	return -1
    }
    set formulaeNode [$window_doc selectNodes formulae]
    if { $formulaeNode ne "" } {

	if { [$formulaeNode @default_database 0] == 1 && \
	    [$formulaeNode @database ""] ne "" } {
	    $lognoter_db sql exec "drop table \"[$formulaeNode @database]\""
	}
	if { $formulaeNode eq [$window_doc documentElement] } {
	    $window_doc delete
	    unset window_doc
	} else {
	    $formulaeNode delete
	}
    }
    create_windowD_do
    return 0
}

# edit properties window
# args: ?-field_name field_name? ?-copy boolean? ?-copy_children boolean?
#       ?-location "before|after field_name"? wp
proc formulae::edit_properties { args } {
    
    package require tile

    set field_name ""
    set copy_field 0
    set copy_children 0
    set location ""
    while { [string match -* [lindex $args 0]] } {
	switch -- [lindex $args 0] {
	    -field_name {
		set args [lrange $args 1 end]
		set field_name [lindex $args 0]
	    }
	    -location {
		set args [lrange $args 1 end]
		set location [lindex $args 0]
	    }
	    -copy {
		set args [lrange $args 1 end]
		set copy_field [lindex $args 0]
	    }
	    -copy_children {
		# this is not still implemented
		set args [lrange $args 1 end]
		set copy_children [lindex $args 0]
		if { $copy_children } { set copy_field 1}
	    }
	    default { error "error in formulae::edit_properties" }
	}
	set args [lrange $args 1 end]
    }
    foreach "wp" $args break

    destroy $wp._props
    set w [dialogwin_snit $wp._props -title [_ "Change properties"] \
	    -callback [list formulae::edit_properties_do] -grab 0 \
	    -morebuttons [list [_ Apply]]]


    $w set_uservar_value refresh_pending 0

    set f [$w giveframe]

################################################################################
#    property type
################################################################################

    set l1 [ttk::labelframe $f.l1 -text [_ "property type"]]
    
    ttk::radiobutton $l1.r1 -text [_ Parameter] -variable [$w give_uservar prop_type] \
	-value param -command [list formulae::edit_properties_update -step prop_type \
	    -original_field_name $w]
    ttk::radiobutton $l1.r2 -text [_ Condition] -variable [$w give_uservar prop_type] \
	-value condition -command [list formulae::edit_properties_update -step prop_type \
	    -original_field_name $w]
    ttk::radiobutton $l1.r3 -text [_ Container] -variable [$w give_uservar prop_type] \
	-value container -command [list formulae::edit_properties_update -step prop_type \
	    -original_field_name $w]

    grid $l1.r1 -sticky nw
    grid $l1.r2 -sticky nw
    grid $l1.r3 -sticky nw

################################################################################
#    navigate
################################################################################

    set l12 [ttk::labelframe $f.l12 -text [_ "navigate"]]

    ttk::button $l12.b1 -text [_ "Previous"] -image [cu::get_icon back-16] -compound left
    $w set_uservar_value previous_button $l12.b1

    ttk::button $l12.b2 -text [_ "Next"] -image [cu::get_icon forward-16] -compound right
    $w set_uservar_value next_button $l12.b2

    ttk::menubutton $l12.mb1 -text [_ "Go to"] -menu $l12.mb1.m
    $w set_uservar_value goto_menu [menu $l12.mb1.m -tearoff 0]

    ttk::menubutton $l12.mb2 -text [_ Action] -menu $l12.mb2.m
    menu $l12.mb2.m -tearoff 0
    $l12.mb2.m add command -label [_ New] -command [list formulae::edit_properties_do \
	    $w 3]
    $l12.mb2.m add command -label [_ Copy] -command [list formulae::edit_properties_do \
	    $w 4]
    $l12.mb2.m add command -label [_ Delete] -command [list formulae::edit_properties_do \
	    $w 5]

    ttk::checkbutton $l12.cb1 -text [_ "Update current"] -variable \
	[$w give_uservar update_current 1]

    ttk::checkbutton $l12.cb2 -text [_ "Refresh"] -variable \
	[$w give_uservar refresh 0]

    grid $l12.b1 $l12.b2 -sticky w -padx 2
    grid $l12.mb1 $l12.mb2 -sticky w
    grid $l12.cb1 $l12.cb2 -sticky w -padx 2

    set l2 [ttk::labelframe $f.l2 -text [_ "properties"]]

################################################################################
#    location
################################################################################

    ttk::label $l2.l0 -text [_ Location:]
    ttk::frame $l2.f0
    ttk::combobox $l2.f0.cb1 -textvariable [$w give_uservar container ""] \
	-state readonly
    bind $l2.f0.cb1 <<ComboboxSelected>> [list formulae::edit_properties_update \
	    -step container -original_field_name $w]

    ttk::combobox $l2.f0.cb2 -textvariable [$w give_uservar before_after [_ After]] \
	-width 7 -state readonly -values [list [_ Before] [_ After]]
    ttk::combobox $l2.f0.cb3 -textvariable [$w give_uservar before_after_name] \
	-state readonly

    ttk::menubutton $l2.f0.m1 -image [cu::get_icon keditbookmarks-16] -style Toolbutton -menu $l2.f0.m1.m
    set m [menu $l2.f0.m1.m -tearoff 0]

    $m add checkbutton -label [_ "Same page"] -variable [$w give_uservar same_page 0]
    
    grid $l2.f0.cb1 $l2.f0.cb2 $l2.f0.cb3 $l2.f0.m1 -sticky w
    grid configure $l2.f0.cb1 $l2.f0.cb3 -sticky ew
    grid columnconfigure $l2.f0 "0 2" -weight 1

################################################################################
#    name & Short name & Print name
################################################################################

    ttk::label $l2.l1 -text [_ Name:]
    ttk::combobox $l2.cb1 -textvariable [$w give_uservar name ""]

    foreach i [list $l2.l1 $l2.cb1] {
	bind $i <1> {
	    %W instate disabled {
		set ret [snit_messageBox -message [_ "Do you want to rename field?"] \
		        -type okcancel -parent %W]
		if { $ret eq "cancel" } { return }
		%W state !disabled
	    }
	}
    }
    ttk::label $l2.l2 -text [_ "Short name:"]
    ttk::combobox $l2.cb2 -textvariable [$w give_uservar short_name ""]
   
    ttk::labelframe $l2.l3
    ttk::label $l2.l3.l3 -text [_ "Print name:"]
    $w set_uservar_value print_name_widget [wordwidget_and_toolbox $l2.l3.wns1 -height 2 \
	    -bd 1 -mode static -scroll 0]
    grid $l2.l3.l3 $l2.l3.wns1 -sticky ew -padx 2 -pady 2
    grid columnconfigure $l2.l3 1 -weight 1

################################################################################
#    expression & condition & field type
################################################################################
   
    set field_types [list numeric text "long text" expression "text expression" \
	    "options editable" "options non-editable" formatted \
	    "formatted expression" button date file]

    set dict [dict create numeric [_ numeric] text [_ text] expression \
	    [_ expression] "text expression" [_ "text expression"] \
	    "options editable" [_ "options editable"] \
	    "options non-editable" [_ "options non-editable"] formatted \
	    [_ formatted] "formatted expression" [_ "formatted expression"] \
	    button [_ button] date [_ date]]

    ttk::frame $l2.f35
    ttk::label $l2.f35.l -text [_ "Field type:"]
    cu::combobox $l2.f35.cb -textvariable [$w give_uservar field_type numeric] \
	-values $field_types -dict $dict \
	-state readonly -width 18
    ttk::frame $l2.f35.f
    ttk::label $l2.f35.f.l -text [_ "Values:"]
    ttk::entry $l2.f35.f.e -textvariable [$w give_uservar valuesList ""]

    set t [_ "Enter a comma separator list of values here"]
    tooltip $l2.f35.f.l $t
    tooltip $l2.f35.f.e $t

    set a "$l2.f35.f.l $l2.f35.f.e" ;# @values
    set e "$l2.f1.m1 $l2.f1.m2" ;# expressions helpers
    set u "$l2.l5 $l2.cb3 $l2.m3" ;# units
    set d [dict create numeric $u text "" "long text" "" \
	    expression "$e $u" "options editable" \
	    $a "options non-editable" $a formatted "" "formatted expression" $e \
	    "text expression" $e]
    $w enable_disable_on_key field_type $d

    set a [list value_label [_ Value:]]
    set b [list value_label [_ Expression:]]
    set c [list value_label [_ Command:]]
    set d [dict create numeric $a text $a "long text" $a \
	    expression $b "options editable" \
	    $a "options non-editable" $a formatted $a "formatted expression" $b \
	    "text expression" $b button $c]
    $w change_key_on_key field_type $d

    grid $l2.f35.f.l $l2.f35.f.e -sticky w
    grid configure $l2.f35.f.e -sticky ew
    grid columnconfigure $l2.f35.f 1 -weight 1
    grid $l2.f35.l $l2.f35.cb $l2.f35.f -sticky w
    grid configure $l2.f35.f -sticky ew
    grid columnconfigure $l2.f35 2 -weight 1

    ttk::label $l2.l4 -textvariable [$w give_uservar value_label]

    set examples {
	123.45
	A+2*F+sin(N2)*sqrt(V3)
	B>=C
    }
    regsub -all {\s*\n\s*} $examples "\n\t" examples
    set examples [string trim $examples]

    tooltip $l2.l4 [_ "Examples:\n\t%s" $examples]
    ttk::frame $l2.f1
    $w set_uservar_value value_widget [ttk::entry $l2.f1.e1 -textvariable \
	    [$w give_uservar value] -width 20]

    ttk::menubutton $l2.f1.m1 -image [cu::get_icon keditbookmarks-16] \
	-style Toolbutton -menu $l2.f1.m1.m
    $w set_uservar_value value_fields_widget [menu $l2.f1.m1.m -tearoff 0]

    ttk::menubutton $l2.f1.m2 -image [cu::get_icon sum-16] -style Toolbutton \
	-menu $l2.f1.m2.m
    menu $l2.f1.m2.m -tearoff 0
    set columnbreak 0

    set funcs [list max(,) min(,) interpolate(,) dictionary(,) pi() --- + - * / ^ ---- \
	    sin() cos() tan() asin() acos() atan() ---- \
	    sinh() cosh() tanh() atan2(,) ---- \
	    abs() sqrt() exp() log() log10() pow(,) hypot(,)]

    set columnbreak 0
    foreach ff1 [formulae eval info commands ::tcl::mathfunc::*] {
	set f1 [namespace tail $ff1]
	if { [lsearch -glob $funcs "${f1}(*"] == -1 } {
	    if { !$columnbreak } {
		lappend funcs ----
		set columnbreak 1
	    }
	    lappend funcs "${f1}()"
	}
    }

    foreach i $funcs {
	if { $i eq "---" } {
	    $l2.f1.m2.m add separator
	} elseif { $i eq "----" } {
	    set columnbreak 1
	} else {
	    $l2.f1.m2.m add command -label $i -command \
		[list formulae::_insert_in_entry $l2.f1.e1 $i] \
		-columnbreak $columnbreak
	    set columnbreak 0
	}
    }
    grid $l2.f1.e1 $l2.f1.m1 $l2.f1.m2 -sticky w
    grid configure $l2.f1.e1 -sticky we
    grid columnconfigure $l2.f1 0 -weight 1

################################################################################
#    units
################################################################################

    ttk::label $l2.l5 -text [_ Units:]
    ttk::combobox $l2.cb3 -textvariable [$w give_uservar units] -values \
	[list m cm mm N kN Pa MPa m\u0b2 m\u0b3 m\u2074]
    ttk::menubutton $l2.m3 -image [cu::get_icon keditbookmarks-16] \
	-style Toolbutton -menu $l2.m3.m
    menu $l2.m3.m -tearoff 0
    foreach i [list \u0b2 \u0b3 \u2074] {
	$l2.m3.m add command -label $i -command [list $l2.cb3 insert insert $i]
    }

################################################################################
#    true & false
################################################################################

    ttk::label $l2.l54 -text [_ "True text:"]
    ttk::combobox $l2.cb31 -textvariable [$w give_uservar true]
    ttk::label $l2.l56 -text [_ "False text:"]
    ttk::combobox $l2.cb32 -textvariable [$w give_uservar false]

################################################################################
#    help
################################################################################

    ttk::label $l2.l6 -text [_ Help]
    $w set_uservar_value help_widget [text $l2.t -wrap word -height 4 -width 20]

################################################################################
#    grid
################################################################################

    grid $l2.l0 $l2.f0 - -sticky w -pady 1
    grid $l2.l1 $l2.cb1 - -sticky w -pady 1
    grid $l2.l2 $l2.cb2 - -sticky w -pady 1
    grid $l2.l3 - - -sticky ew -pady 1 -padx 2
    grid $l2.f35 - - -sticky ew -pady 1
    grid $l2.l4 $l2.f1 - -sticky w -pady 1
    grid $l2.l5 $l2.cb3 $l2.m3 -sticky w -pady 1
    grid $l2.l54 $l2.cb31 -sticky w -pady 1
    grid $l2.l56 $l2.cb32 -sticky w -pady 1
    grid $l2.l6 - - -sticky w  -padx "20 0" -pady 1
    grid $l2.t  - - -sticky wens -padx "20 4" -pady "1 4"

    grid configure $l2.f0 $l2.cb1 $l2.cb2 $l2.f1 $l2.cb3 $l2.cb31 $l2.cb32 \
	-sticky ew -padx "0 2"
    grid columnconfigure $l2 1 -weight 1
    grid rowconfigure $l2 10 -weight 1

    grid $l1 $l12 -sticky nsew
    grid $l2  - -sticky nsew

    grid configure $l12 -padx "2 0"
    grid columnconfigure $f 0 -weight 1
    grid rowconfigure $f 1 -weight 1

    tk::TabToWindow $l2.cb1
    bind $w <Return> [list $w invokeok]
    
    edit_properties_update -field_name $field_name -copy $copy_field \
	-location $location $w
    $w createwindow
}

proc formulae::_insert_in_entry { entry text } {
    focus $entry
    catch {
	set insert [$entry index insert]
	if {([$entry index sel.first] <= $insert)
	    && ([$entry index sel.last] >= $insert)} {
	    $entry delete sel.first sel.last
	}
    }
    set i [$entry index insert]
#     if { $i == 0 } {
#         $entry icursor end
#         set i [$entry index insert]
#     }
    $entry insert insert $text
    set j [string first "(" $text]
    if { $j != -1 } {
	$entry icursor [expr {$i+$j+1}]
    }
}

proc formulae::edit_properties_do { w { action "" } } {

    if { $action eq "" } {
	set action [$w giveaction]
    }

    switch -- $action {
	-1 - 0 {
	    if { [$w give_uservar_value refresh_pending] } {
		create_windowD_do
	    }
	}
	1 {
	    set err [catch { edit_properties_accept $w } errstring]
	    if { $err } {
		snit_messageBox -icon error -message $errstring
		#snit_messageBox -icon error -message $::errorInfo
		return
	    }
	}
	2 {
	    set err [catch { edit_properties_accept $w } errstring]
	    if { $err } {
		snit_messageBox -icon error -message $errstring
		return
	    }
	    return
	}
	3 {
	    if { [$w give_uservar_value update_current] && [$w give_uservar_value name] ne "" } {
		set refresh [$w give_uservar_value refresh]
		set err [catch { edit_properties_accept -refresh $refresh $w } errstring]
		if { $err } {
		    snit_messageBox -icon error -message $errstring
		    return
		}
	    }
	    set field_name [$w give_uservar_value name]
	    edit_properties_update -location [list after $field_name] $w
	    return
	}
	4 {
	    set field_name [$w give_uservar_value name]
	    edit_properties_update -original_field_name -copy 1 \
		-location [list after $field_name] $w
	    return
	}
	5 {
	    set refresh [$w give_uservar_value refresh]
	    set err [catch { edit_properties_delete -refresh $refresh $w } errstring]
	    if { $err } {
		snit_messageBox -icon error -message $errstring
		return
	    }
	    edit_properties_update $w
	    return
	}
    }
    destroy $w
}

# what can be: prev, next this
proc formulae::edit_properties_navigate { w ref_field_name what } {
    variable window_doc

    if { [$w give_uservar_value update_current] && [$w give_uservar_value name] ne "" } {
	set refresh [$w give_uservar_value refresh]
	set err [catch { edit_properties_accept -refresh $refresh $w } errstring]
	if { $err } {
	    snit_messageBox -icon error -message $errstring
	    return
	}
    }
    if { $ref_field_name eq "" } {
	set node ""
	if { [$w give_uservar_value name] ne "" } {
	    set name [$w give_uservar_value name]
	    set xp {
		//param[@n='N'] |
		//condition[@n='N'] |
		//container[@n='N']
	    }
	    set xp [string map [list N [xml [$w give_uservar_value name]]] $xp]
	    set node [$window_doc selectNodes $xp]
	    if { $node ne "" } {
		set ref_field_name [$node @n]
	    }
	}
	if { $node eq "" } {
	    set nodes [$window_doc selectNodes "//param|//condition|//container"]
	    set node [lindex $nodes 0]
	    if { $node ne "" } {
		set ref_field_name [$node @n]
		#             if { [llength $nodes] == 1 } {
		    #                 set what this
		    #             } elseif { $what eq "prev" } { set what next }
	    } else {
		set what new
	    }
	}
    }
    set xp {
	//param[@n='N'] |
	//condition[@n='N'] |
	//container[@n='N']
    }
    set xp [string map [list N $ref_field_name] $xp]

    set ref_node [$window_doc selectNodes $xp]
    switch $what {
	prev {
	    set nodes [$window_doc selectNodes //param|//condition|//container]
	    set ipos [lsearch -exact $nodes $ref_node]
	    set node [lindex $nodes [incr ipos -1]]
	}
	next {
	    set nodes [$window_doc selectNodes //param|//condition|//container]
	    set ipos [lsearch -exact $nodes $ref_node]
	    set node [lindex $nodes [incr ipos 1]]
	}
    }
    if { $node ne "" } {
	set name [$node @n ""]
	if { $name eq "" } { set name [$node @value] }
	set location ""
    } else {
	set name ""
	switch $what prev { set location before }  next { set location after }
	lappend location $ref_field_name
    }
    edit_properties_update -field_name $name -location $location $w
}

proc formulae::defaultvalue { root name } {
    return [$root selectNodes "string(//param\[@n='[xml $name]'\]/@defaultvalue)"]
}

proc formulae::mathfunc_max { args } {
    set res ""
    foreach i $args {
	if { $res eq "" || $i > $res } {
	    set res $i
	}
    }
    return $res
}

proc formulae::mathfunc_min { args } {
    set res ""
    foreach i $args {
	if { $res eq "" || $i < $res } {
	    set res $i
	}
    }
    return $res
}

proc formulae::mathfunc_case { args } {
    set res ""
    foreach "cond value" $args {
	if { $value eq "" } {
	    set value $cond
	    set cnd 1
	} else {
	    set cnd [uplevel 1 expr $cond]
	}
	if { $cnd } { return [uplevel 1 expr $value] }
    }
    return 0.0
}

proc formulae::mathfunc_pi { args } {
    ::math::constants::constants pi pi
    return $pi
}

proc formulae::mathfunc_interpolate { val args } {
    if { $val < [lindex $args 0] } {
	return [lindex $args 1]
    }
    set lenm2 [expr {[llength $args]-2}]
    for { set i 0 } { $i < $lenm2 } { incr i 2 } {
	set x [lindex $args $i]
	set xnext [lindex $args [expr {$i+2}]]
	if { $val >= $x && $val <= $xnext } {
	    set deltax [expr {($val-$x)/double($xnext-$x)}]
	    return [expr {(1.0-$deltax)*[lindex $args [expr {$i+1}]]+\
		    $deltax*[lindex $args [expr {$i+3}]]}]
	}
    }
    return [lindex $args end]
}

proc formulae::mathfunc_dictionary { val args } {
    return [dict get $args $val]
}


proc formulae::give_updated_tree {} {
    variable window_doc
    variable uvalues
    variable current_set_name
    variable current_set_print

    window_actualize -exec_callback 0

    if { ![info exists window_doc] } { return "" }
    set root [$window_doc documentElement]
    if { $root eq "" } { return "" }

    set current_set_name [string trim $current_set_name]
    set secnameNode [$root selectNodes .//setname]
    if { $secnameNode ne "" && $current_set_name ne "" } {
	set cnode [$secnameNode selectNodes "name\[@n='[xml $current_set_name]'\]"]
	if { $cnode ne "" } { $cnode delete }
	set cnode [$secnameNode selectNodes {name[1]}]
	set newnode [[$secnameNode ownerDocument] createElement name]
	$newnode setAttribute n $current_set_name print $current_set_print
	if { $cnode ne "" } {
	    $secnameNode insertBefore $newnode $cnode
	} else {
	    $secnameNode appendChild $newnode
	}
	$secnameNode setAttribute n $current_set_name
    }
    foreach i [array names uvalues] {
	regexp {,(.*)} $i {} n
	set node [$root selectNodes "//param\[@n='[xml $n]'\]"]
	if { [$node @dict ""] ne "" } {
	    set v [dict get [dict_inverse [$node @dict]] $uvalues($i)]
	} else {
	    set v $uvalues($i)
	}
	$node setAttribute value $v
	if { [info exists newnode] } {
	    set np [$newnode appendChildTag name_param]
	    $np setAttribute n $n value $v
	}
    }
    return $window_doc
}

proc formulae::give_params_dict {} {
    variable window_doc
    variable uvalues

    if { ![info exists window_doc] } { return "" }
    set root [$window_doc documentElement]
    if { $root eq "" } { return "" }

    set d ""
    foreach i [array names uvalues] {
	regexp {,(.*)} $i {} n
	set node [$root selectNodes "//param\[@n='[xml $n]'\]"]
	if { [$node @dict ""] ne "" } {
	    set dict [dict_inverse [split [$node @dict] ,]]
	    if { [dict exists $dict $uvalues($i)] } {
		set v [dict get $dict $uvalues($i)]
	    } else {
		set v $uvalues($i)
	    }
	} else {
	    set v $uvalues($i)
	}
	dict set d $n $v
    }
    return $d
}

proc formulae::manage_names { what key combo node } {
    variable uvalues
    variable current_set_name
    variable current_set_print
    variable current_set_print_all_none

    set current_set_name [string trim $current_set_name]

    set cnode [$node selectNodes "name\[@n='[xml $current_set_name]'\]"]

    switch $what {
	"accept" {
	    if { $cnode ne "" } { $cnode delete }
	    set cnode [$node selectNodes {name[1]}]
	    set newnode [[$node ownerDocument] createElement name]
	    $newnode setAttribute n $current_set_name print $current_set_print
	    if { $cnode ne "" } {
		$node insertBefore $newnode $cnode
	    } else {
		$node appendChild $newnode
	    }
	    foreach i [array names uvalues] {
		regexp {,(.*)} $i {} n
		set np [$newnode appendChildTag name_param]
		$np setAttribute n $n value $uvalues($i)
	    }
	    set values ""
	    foreach cnode [$node selectNodes name] {
		lappend values [$cnode @n]
	    }
	    $combo configure -values $values
	}
	"select" {
	    if { $cnode eq "" } { return }
	    foreach i [array names uvalues] {
		regexp {,(.*)} $i {} n
		set pnode [$cnode selectNodes "name_param\[@n='[xml $n]'\]"]
		if { $pnode ne "" } {
		    set uvalues($i) [$pnode @value]
		}
	    }
	    set current_set_print [$cnode @print 1]
	    window_actualize
	}
	"delete" {
	    if { $cnode eq "" } { return }
	    $cnode delete
	    set values ""
	    foreach cnode [$node selectNodes name] {
		lappend values [$cnode @n]
	    }
	    $combo configure -values $values
	    set current_set_name ""
	    set current_set_print 1
	}
	"print" {
	    if { $cnode eq "" } { return }
	    $cnode setAttribute print $current_set_print

	    foreach "numprinted numtotal" [list 0 0] break
	    foreach cnode [$node selectNodes name] {
		incr numprinted [$cnode @print 1]
		incr numtotal
	    }
	    if { $numprinted == $numtotal } {
		set current_set_print_all_none all
	    } elseif { $numprinted > 0 } {
		set current_set_print_all_none some
	    } else {
		set current_set_print_all_none none
	    }
	}
	"printall" {
	    foreach cnode [$node selectNodes name] {
		$cnode setAttribute print 1
	    }
	    set current_set_print 1
	    set current_set_print_all_none all
	}
	"printnone" {
	    foreach cnode [$node selectNodes name] {
		$cnode setAttribute print 0
	    }
	    set current_set_print 0
	    set current_set_print_all_none none
	}
    }
}

proc formulae::destroy_window {} {
    variable uvalues
    variable uvalues_moredata
    variable update_stack
    variable svg_stack
    variable nicer_name
    variable update_callback
    variable window_doc

    set update_stack ""
    set svg_stack ""
    array unset uvalues
    array unset uvalues_moredata
    array unset nicer_name
    set update_callback ""
    if { [info exists window_doc] } {
	$window_doc delete
	unset window_doc
    }
}

snit::widgetadaptor snit_notebook {
    option -lognoter_db ""
    option -database ""
    option -rootnode ""
    option -dbstate normal ;# normal or disabled

    typevariable type_name "notebook"
    variable tabs ""
    variable tab_options
    variable current_tab ""
    variable grid_pos ""
    variable grid_opts ""
    variable search_state 0
    variable is_created 0

    delegate method * to hull
    delegate method paned_add to hull as add
    delegate option * to hull

    constructor args {

	installhull [ttk::paned $win -orient vertical]
	ttk::frame $win.f
	$win paned_add $win.f -weight 0
	#grid propagate $win.f 0

	if { $type_name eq "arrows_small" } { set type_name "arrows_small_h" }

	if { $type_name eq "arrows_small_h" || $type_name eq "arrows_small_v" } {
	    ttk::button $win.f.b1 -image [cu::get_icon start-16] -style Toolbutton \
		-command [mymethod _pn_select start]
	    tooltip $win.f.b1 [_ "Start page"]
	    ttk::button $win.f.b2 -image [cu::get_icon back-16] -style Toolbutton \
	       -command [mymethod _pn_select previous]
	    tooltip $win.f.b2 [_ "Previous page"]
	    ttk::button $win.f.b3 -image [cu::get_icon forward-16] -style Toolbutton \
		-command [mymethod _pn_select next]
	    tooltip $win.f.b3 [_ "Next page"]
	    ttk::button $win.f.b4 -image [cu::get_icon finish-16] -style Toolbutton \
		-command [mymethod _pn_select last]
	    tooltip $win.f.b4 [_ "Last page"]
	    ttk::menubutton $win.f.b5 -text [_ "Go"] -style Toolbutton \
		-menu $win.f.b5.m
	    menu $win.f.b5.m -tearoff 0 -postcommand [mymethod _fill_go_menu $win.f.b5.m]

	    ttk::separator $win.f.sep
	} elseif { $type_name eq "arrows" } {
	    set grid_pos [list -row 2 -column 0]
	    set grid_opts [list -row 2 -column 0 -columnspan 5 -rowspan 1]

	    ttk::button $win.f.b1 -text [_ "Start"] -image [cu::get_icon start-16] \
		-compound left -width 8 -command [mymethod _pn_select start]
	    ttk::button $win.f.b2 -text [_ "Previous"] -image [cu::get_icon back-16] \
		-compound left -width 8 -command [mymethod _pn_select previous]
	    ttk::button $win.f.b3 -text [_ "Next"] -image [cu::get_icon forward-16] \
		-compound right -width 8 -command [mymethod _pn_select next]
	    ttk::button $win.f.b4 -text [_ "Last"] -image [cu::get_icon finish-16] \
		-compound right -width 8 -command [mymethod _pn_select last]
	    ttk::menubutton $win.f.b5 -text [_ "Go"] \
		-menu $win.f.b5.m
	    menu $win.f.b5.m -tearoff 0 -postcommand [mymethod _fill_go_menu $win.f.b5.m]

	    ttk::separator $win.f.sep
	} else {
	    ttk::notebook $win.f.nb
	    ttk::separator $win.f.sep
	}
	$self configurelist $args
	bind $win <Configure> [mymethod _compute_size]

	set is_created 1
	$self _configure_database
    }
    method _grid_buttons {} {
	if { $type_name eq "arrows_small_v" } {
	    set grid_pos [list -row 0 -column 1]
	    set grid_opts [list -row 0 -column 1 -columnspan 1 -rowspan 6]

	    if { [llength $tabs] > 1 } {
		if { $options(-database) eq "" } {
		    set idx 0
		    foreach i [list $win.f.b1 $win.f.b2 $win.f.b3 $win.f.b4 $win.f.sep $win.f.b5] {
		        grid $i -sticky n -row $idx -column 0
		        grid rowconfigure $win.f $idx -weight 0
		        incr idx
		    }
		    grid configure $win.f.b1 -pady "40 0"
		    grid configure $win.f.sep -sticky new -padx 2
		    grid columnconfigure $win.f 1 -weight 1
		    grid rowconfigure $win.f 5 -weight 1
		    if { [winfo exists $win.f.db] } {
		        grid remove $win.f.db
		    }
		    dict set grid_opts -rowspan 6
		} else {
		    set idx 0
		    foreach i [list $win.f.b1 $win.f.b2 $win.f.b3 $win.f.sep $win.f.b5 \
		            $win.f.db] {
		        grid $i -sticky n -row $idx -column 0
		        grid rowconfigure $win.f $idx -weight 0
		        incr idx
		    }
		    grid configure $win.f.b1 -pady "40 0"
		    grid configure $win.f.sep -sticky new -padx 2
		    grid columnconfigure $win.f 1 -weight 1
		    grid rowconfigure $win.f 5 -weight 1
		    grid remove $win.f.b4
		    dict set grid_opts -rowspan 6
		}
	    } else {
		grid remove $win.f.b1 $win.f.b2 $win.f.b3 $win.f.b4 \
		    $win.f.b5 $win.f.sep
		if { $options(-database) eq "" } {
		    if { [winfo exists $win.f.db] } {
		        grid remove $win.f.db
		    }

		} else {
		    set idx 0
		    foreach i [list $win.f.db] {
		        grid $i -sticky n -row $idx -column 0
		        incr idx
		    }
		    grid columnconfigure $win.f 1 -weight 1
		    grid rowconfigure $win.f 0 -weight 1
		    dict set grid_opts -rowspan 1
		}
	    }
	} elseif { $type_name eq "arrows" || $type_name eq "arrows_small_h"} {
	    set grid_pos [list -row 2 -column 0]
	    set grid_opts [list -row 2 -column 0 -columnspan 5 -rowspan 1]
	    if { [llength $tabs] > 1 } {
		if { $options(-database) eq "" } {
		    set idx 0
		    foreach i [list $win.f.b1 $win.f.b2 $win.f.b3 $win.f.b4 $win.f.b5] {
		        grid $i -sticky w -row 0 -column $idx
		        grid columnconfigure $win.f $idx -weight 0
		        incr idx
		    }
		    grid $win.f.sep -padx "0 50" -sticky ew -pady 2 -row 1 -column 0 \
		        -columnspan 5
		    grid columnconfigure $win.f 4 -weight 1
		    if { [winfo exists $win.f.db] } {
		        grid remove $win.f.db
		    }
		    dict set grid_opts -columnspan 5
		} else {
		    set idx 0
		    if { $type_name eq "arrows" } {
		        set list [list $win.f.b1 $win.f.b2 $win.f.b3 $win.f.b5 \
		                $win.f.db]
		    } else {
		        set list [list $win.f.b1 $win.f.b2 $win.f.b3 $win.f.b4 \
		                $win.f.b5 $win.f.db]
		    }
		    foreach i $list {
		        grid $i -sticky w -row 0 -column $idx
		        grid columnconfigure $win.f $idx -weight 0
		        incr idx
		    }
		    grid configure $win.f.db -padx "20 0"
		    grid $win.f.sep -padx "0 50" -sticky ew -pady 2 -row 1 -column 0 \
		        -columnspan [llength $list]
		    grid columnconfigure $win.f [expr {[llength $list]-1}] -weight 1
		    dict set grid_opts -columnspan [llength $list]
		}
		grid rowconfigure $win.f 2 -weight 1
	    } else {
		grid remove $win.f.b1 $win.f.b2 $win.f.b3 $win.f.b4 \
		    $win.f.b5 $win.f.sep
		if { $options(-database) eq "" } {
		    if { [winfo exists $win.f.db] } {
		        grid remove $win.f.db
		    }
		} elseif { [winfo exists $win.f.db] } {
		    set idx 0
		    foreach i [list $win.f.db] {
		        grid $i -sticky w -row 0 -column $idx
		        grid columnconfigure $win.f $idx -weight 0
		        incr idx
		    }
		    grid $win.f.sep -padx "0 50" -sticky ew -pady 2 -row 1 -column 0 \
		        -columnspan 1
		    grid columnconfigure $win.f 0 -weight 1
		    grid rowconfigure $win.f 2 -weight 1
		    dict set grid_opts -columnspan 1
		}
	    }
	} else {
	    if { $options(-database) eq "" } {
		grid $win.f.nb -sticky nsew -row 0 -column 0
		grid columnconfigure $win.f 0 -weight 1
		grid rowconfigure $win.f 0 -weight 1
		if { [winfo exists $win.f.db] } {
		    grid remove $win.f.db
		}
	    } else {
		set idx 0
		foreach i [list $win.f.db] {
		    grid $i -sticky w -row 0 -column $idx
		    grid columnconfigure $win.f $idx -weight 0
		    incr idx
		}
		grid $win.f.sep -padx "0 50" -sticky ew -pady 2 -row 1 -column 0 \
		    -columnspan 1
		grid $win.f.nb -row 1 -column 0 -sticky nsew -pady 2 -row 2 -column 0 \
		    -columnspan 1
		grid columnconfigure $win.f 0 -weight 1
		grid rowconfigure $win.f 2 -weight 1
		dict set grid_opts -columnspan 1
		dict set grid_opts -row 1
		dict set grid_pos -row 1
	    }
	}
    }
    method _configure_database {} {
	if { !$is_created } { return }

	if { $options(-lognoter_db) ne "" && $options(-database) ne "" && \
	     $options(-rootnode) ne "" } {
	    if { ![winfo exists $win.list] } {
		set db $options(-lognoter_db)
		set table $options(-database)
		set columns ""
		lappend columns [list 6 id left text 1]
		set cols [$db sql column_names $table]
		foreach col [lrange $cols 3 end] {
		    if { [regexp {__(contents|cdate|mdate|type|size|path)$} $col] } {
		        continue
		    }
		    lappend columns [list 14 $col left text 1]
		}
		package require fulltktree
		fulltktree $win.list -columns $columns -showbuttons 0 -showlines 0 \
		    -bd 2 -relief raised -sensitive_cols all \
		    -contextualhandler [mymethod manage_database contextual] \
		    -selecthandler2 [mymethod manage_database fill_entries] \
		    -deletehandler [mymethod manage_database delete]
	    }
	    $win paned_add $win.list -weight 1
	    $self manage_database update

	    if { ![winfo exists $win.f.db] } {
		ttk::labelframe $win.f.db -style WNtoolbar
		ttk::button $win.f.db.create -image [cu::get_icon filenew-16] \
		    -style WNbutton \
		    -command [mymethod manage_database create]
		ttk::button $win.f.db.edit -image [cu::get_icon edit-16] \
		    -style WNbutton \
		    -command [mymethod manage_database edit]
		cu::menubutton_checkbutton $win.f.db.find -image [cu::get_icon kfind-16] -style Toolbutton \
		    -variable [myvar search_state] -menu $win.f.db.find.m \
		    -command [mymethod manage_database search] \
		    -style WNbutton

		menu $win.f.db.find.m -tearoff 0
		$win.f.db.find.m add command -image [cu::get_icon kfind-16] \
		    -label [_ "Search in field"] -compound left \
		    -command [mymethod manage_database search]
		$win.f.db.find.m add command -label [_ "Search in any field"] \
		    -command [mymethod manage_database search_any]

		ttk::button $win.f.db.delete -image [cu::get_icon remove-16] \
		    -style WNbutton \
		    -command [mymethod manage_database delete]

		grid $win.f.db.create $win.f.db.edit $win.f.db.find $win.f.db.delete \
		    -sticky w
		grid columnconfigure $win.f.db 3 -weight 1

		tooltip $win.f.db.create [_ "Create new row"]
		tooltip $win.f.db.edit [_ "Modify selected row"]
		tooltip $win.f.db.find [_ "Find rows by entering text in fields"]
		tooltip $win.f.db.delete [_ "Delete selected rows"]
	    }
	    if { $options(-dbstate) eq "disabled" } {
		foreach i [list $win.f.db.create $win.f.db.edit $win.f.db.find \
		       $win.f.db.delete] {
		    $i state disabled
		} 
	    }
	    $self _grid_buttons
	} elseif { $options(-lognoter_db) eq "" && $options(-database) eq "" } {
	    if { [winfo exists $win.list] } {
		destroy $win.list
	    }
	    $self _grid_buttons
	}
    }
    onconfigure -database {value} {
	set options(-database) $value
	$self _configure_database
    }
    onconfigure -lognoter_db {value} {
	set options(-lognoter_db) $value
	$self _configure_database
    }
    onconfigure -rootnode {value} {
	set options(-rootnode) $value
	$self _configure_database
    }
    onconfigure -dbstate {value} {
	set options(-dbstate) $value
	$self _configure_database
    }
    typemethod notebook_type { _type_name } {
	set type_name $_type_name
    }
    method add { w args } {
	set tab_options([llength $tabs]) $args
	if { ![dict exists $tab_options([llength $tabs]) -state] } {
	    dict set tab_options([llength $tabs]) -state normal
	}
	lappend tabs $w
	if { $type_name eq "notebook" } {
	    eval [list $win.f.nb add $w] $args
	} else {
	    if { [llength $tabs] == 1 } {
		$self select 0
	    } else {
		$win.f.b3 state !disabled
		$win.f.b4 state !disabled
	    }
	}
	$self _grid_buttons
    }
    method tab { index args } {
	if { $type_name eq "notebook" } {
	    eval [list $win.f.nb tab $index] $args
	} else {
	    set index [$self index $index]
	    if { ![llength $args] } {
		return $tab_options($index)
	    } elseif { [llength $args] == 1 } {
		if { [dict exists $tab_options($index) [lindex $args 0]] } {
		    return [dict get $tab_options($index) [lindex $args 0]]
		} else { return "" }
	    } else {
		foreach "n v" $args { dict set tab_options($index) $n $v }
	    }
	}
    }
    method index { index } {
	if { $type_name eq "notebook" } {
	    return [$win.f.nb index $index]
	} else {
	    if { [set ipos [lsearch -exact $tabs $index]] != -1 } {
		return $ipos
	    }
	    # warning: end could be llength-1
	    set index [string map [list current $current_tab \
		        end [llength $tabs]] $index]
	   return [expr $index]
	}
    }
    proc _add_to_bindtag_recursive { w tag } {
	if { [lsearch [bindtags $w] $tag] == -1 } {
	    bindtags $w [concat $tag [bindtags $w]]
	}
	foreach wc [winfo children $w] {
	    _add_to_bindtag_recursive $wc $tag
	}
    }
    method select { index } {

	set index [$self index $index]
	set w [lindex $tabs $index]
	_add_to_bindtag_recursive $w notebook_children

	if { $type_name eq "notebook" } {
	    $win.f.nb select $index
	} else {
	    set old_w [eval [list grid slaves $win.f] $grid_pos]
	    if { $old_w ne "" } { grid remove $old_w }
	    eval [list grid $w -in $win.f -sticky nsew] $grid_opts
	    set current_tab $index
	    for { set i 1 } { $i <= 4 } { incr i } {
		$win.f.b$i state !disabled
	    }
	    if { $index == 0 } {
		$win.f.b1 state disabled
		$win.f.b2 state disabled
	    }
	    if { $index == [llength $tabs]-1 } {
		$win.f.b3 state disabled
		$win.f.b4 state disabled
	    }
	}
    }
    method _compute_size {} {
	if { $type_name eq "notebook" } { return }
	update idletasks

	if { $type_name eq "arrows_small" } {
	    foreach "- - base_width base_height" [grid bbox $win.f 0 0 1 0] break
	    #set base_height 0
	} else {
	    set base_width 0
	    foreach "- - - base_height" [grid bbox $win.f 0 0 0 1] break
	}
	foreach i $tabs {
	    set index [$self index $i]
	    if { [dict get $tab_options($index) -state] eq "hidden" } { continue }
	    if { [winfo reqwidth $i]+$base_width > [$win.f cget -width] } {
		$win.f configure -width [expr {[winfo reqwidth $i]+$base_width}]
	    }
	    if { [winfo reqheight $i]+$base_height > [$win.f cget -height] } {
		$win.f configure -height [expr {[winfo reqheight $i]+$base_height}]
	    }
	}
    }
    method _pn_select { what } {
	switch $what {
	    previous {
		set tab [expr {$current_tab-1}]
		while { $tab >= 0 && [$self tab $tab -state] ne "normal" } {
		    incr tab -1
		}
		if { $tab < 0 } { bell ; return }
		$self select $tab
	    }
	    next {
		set len [llength $tabs]
		set tab [expr {$current_tab+1}]
		while { $tab < $len && [$self tab $tab -state] ne "normal" } {
		    incr tab 1
		}
		if { $tab >= $len } { bell ; return }
		$self select $tab
	    }
	    start {
		set len [llength $tabs]
		set tab 0
		while { $tab < $len && [$self tab $tab -state] ne "normal" } {
		    incr tab 1
		}
		if { $tab >= $len } { bell ; return }
		$self select $tab
	    }
	    last {
		set tab [expr {[llength $tabs]-1}]
		while { $tab >= 0 && [$self tab $tab -state] ne "normal" } {
		    incr tab -1
		}
		if { $tab < 0 } { bell ; return }
		$self select $tab
	    }
	}
	$self _compute_size
    }
    method _fill_go_menu { menu } {
	$menu delete 0 end
	for { set i 0 } { $i < [llength $tabs] } { incr i } {
	    if { [dict get $tab_options($i) -state] eq "hidden" } { continue }
	    if { $i == $current_tab } {
		$menu add checkbutton -command [mymethod select $i] \
		    -variable kk
		set ::kk 1
	    } else {
		$menu add command -command [mymethod select $i]
	    }
	    foreach "v l" [list -text -label -image -image -compound -compound \
		    -state -state] {
		if { [dict exists $tab_options($i) $v] } {
		    $menu entryconfigure end $l [dict get $tab_options($i) $v]
		}
	    }
	}
	if { [winfo exists $win.list] } {
	    $menu add separator
	    $menu add command -label [_ List] -command [mymethod viewlist]
	}

    }
    method viewlist {} {
	$win sashpos 0 0
    }
    variable _clear_save_entries_recursive_dict
    method _clear_save_entries_recursive { w } {

	set err [catch { $w get } content]
	if { $err } {
	    set err [catch { $w get 1.0 end-1c } content]
	}
	if { !$err } {
	    dict set _clear_save_entries_recursive_dict $w $content
	    catch { $w delete 0 end }
	    catch { $w ClearText }
	}
	foreach wc [winfo children $w] {
	    $self _clear_save_entries_recursive $wc
	}
    }
    method _restore_entries_recursive { w } {

	if { [dict exists $_clear_save_entries_recursive_dict $w] } {
	    set v [dict get $_clear_save_entries_recursive_dict $w]
	    catch {
		$w delete 0 end
		$w insert end $v
	    }
	    catch {
		$w ClearText
		$w ApplyXMLdataAsText $v
	    }
	}
	foreach wc [winfo children $w] {
	    $self _restore_entries_recursive $wc
	}
    }
    variable db_update_idle
    # manage database
    method manage_database { args } {
	set db $options(-lognoter_db)
	set table $options(-database)

	set optional {
	}
	set compulsory "what"
	set args [parse_args -raise_compulsory_error 0 $optional \
		$compulsory $args]
	
	switch $what {
	    create - edit {
		if { $what eq "edit" } {
		    set items [$win.list selection get]
		    if { [llength $items] == 0 } {
		        snit_messageBox -message \
		            [_ "It is necessary to select one row in order to modify it"] \
		            -parent $win
		        return
		    }
		    if { [llength $items] > 1 } {
		        snit_messageBox -message [_ "It is necessary to select only one row in order to modify it"] \
		            -parent $win
		        return
		    }
		    set item [lindex $items 0]
		    set id [$win.list item text $items 0]
		} else {
		    set id ""
		    set item end
		}
		set cols [$db sql column_names $table]
		set cols_r [lrange $cols 3 end]
		set valuesList [lrepeat [expr {[llength $cols]-3}] null]
		foreach "values values_moredata" \
		    [formulae::window_actualize -exec_callback 0] break
		foreach "n v" $values {
		    set ipos [lsearch -exact $cols_r $n]
		    if { $ipos == -1 } { continue }
		    set v [string trim $v]

		    set xp [format_xpath {string(//param[@n=%s]/@field_type)} $n]
		    set field_type [$options(-rootnode) selectNodes $xp]
		    if { $field_type eq "file" } {
		        set err [catch { dict get $values_moredata $n contents } c]
		        if { $err || $c eq "" } {
		            set v ""
		        } elseif { $v eq "" } {
		            dict set values_moredata $n ""
		        }
		    }
		    if { [string is double -strict $v] } {
		        lset valuesList $ipos [$db sql escape $v]
		    } else {
		        lset valuesList $ipos '[$db sql escape $v]'
		    }
		}
		foreach "n v" $values_moredata {
		    dict for "n2 v2" $v {
		        set nfull "${n}__$n2"
		        set ipos [lsearch -exact $cols_r $nfull]
		        if { $ipos == -1 } { continue }
		        lset valuesList $ipos '[$db sql escape $v2]'
		    }
		}
		if { $id eq "" } {
		    set cmd "insert into \"$table\" (\"[join $cols \",\"]\") values(
		        null,now(),now(),"
		    append cmd [join $valuesList ,] ")"
		} else {
		    set cmd "update \"$table\" set [lindex $cols 2]=now()"
		    foreach n [lrange $cols 3 end] v $valuesList {
		        append cmd ",\"$n\"=$v"
		    }
		    append cmd " where id=$id"
		}
		$db sql exec $cmd
		$self manage_database update
	    }
	    fill_entries {
		set items [$win.list selection get]
		if { [llength $items] == 0 && [$win.list item count] == 2 } {
		    $win.list selection add "first visible"
		    set items [$win.list selection get]
		}
		if { [llength $items] == 0 } {
		    snit_messageBox -message [_ "It is necessary to select one row in order to modify it"] \
		        -parent $win
		    return
		}
		if { [llength $items] > 1 } {
		    snit_messageBox -message [_ "It is necessary to select only one row in order to modify it"] \
		        -parent $win
		    return
		}
		set id [$win.list item text $items 0]

		set cols [$db sql column_names $table]
		set values [$db sql sel "select * from $table 
		        where id=$id"]

		foreach "input_values input_values_more" [list "" ""] break
		set idx 3
		foreach col [lrange $cols 3 end] {
		    if { [regexp {(.*)__(.*)} $col {} name subname] } {
		        dict set input_values_more $name $subname [lindex $values $idx]
		    } else {
		        dict set input_values $col [lindex $values $idx]
		    }
		    incr idx
		}
		formulae::fill_form $input_values $input_values_more

		if { $search_state eq "any" } {
		    $self manage_database search_any
		}
		if { $search_state == 1 } {
		    set search_state 0
		    $self manage_database search
		}
	    }
	    delete {
		set items [$win.list selection get]
		if { [llength $items] == 0 } {
		    snit_messageBox -message [_ "It is necessary to select rows in order to delete them"] \
		        -parent $win
		    return
		}
		set text [_ "Are you sure to delete %d rows?" [llength $items]]
		set retval [snit_messageBox -default cancel -icon question -message $text \
		        -type okcancel -parent $win]
		if { $retval == "cancel" } { return }
		$db sql exec "begin"
		foreach item $items {
		    set id [$win.list item text $item 0]
		    $db sql exec "delete from \"$table\" where id=$id"
		}
		$db sql exec "commit"
		$self manage_database update
	    }
	    search_any {
		if { $search_state == 1 } {
		    set search_state 0
		    $self manage_database search
		}

		if { $search_state eq "any" } {
		    $win sashpos 0 [expr {[$win sashpos 0]- \
		            [winfo reqheight $win.f.db.searchframe]}]
		    destroy $win.f.db.searchframe
		    set search_state 0
		    $self manage_database update
		    return
		}
		set search_state any
		ttk::labelframe $win.f.db.searchframe -text [_ search] \
		    -style WNtoolbar -padding 2
		ttk::label $win.f.db.searchframe.l -text [_ "Search"]: \
		    -style WNbutton
		ttk::entry $win.f.db.searchframe.e
		ttk::button $win.f.db.searchframe.b1 -image [cu::get_icon kfind-16] \
		    -style WNbutton \
		    -command [mymethod manage_database update_idle]
		ttk::button $win.f.db.searchframe.b2 -image [cu::get_icon remove-16] \
		    -style WNbutton \
		    -command [mymethod manage_database search_any]

		tooltip $win.f.db.searchframe.b1 [_ "Search text"]
		tooltip $win.f.db.searchframe.b2 [_ "Close search window"]

		bind $win.f.db.searchframe.e <KeyPress> [mymethod manage_database update_idle]
		bind $win.f.db.searchframe.e <Escape> [mymethod manage_database search_any]
		bind $win.f.db.searchframe.e <Return> [mymethod manage_database fill_entries]

		grid $win.f.db.searchframe.l $win.f.db.searchframe.e \
		    $win.f.db.searchframe.b1 $win.f.db.searchframe.b2 -sticky w -pady 1
		grid configure $win.f.db.searchframe.l -padx 2
		grid $win.f.db.searchframe -row 1 -column 0 -columnspan 4 \
		    -padx 2 -pady 2
		focus $win.f.db.searchframe.e
		update idletasks

		$win sashpos 0 [expr {[$win sashpos 0]+ \
		        [winfo reqheight $win.f.db.searchframe]}]
	    }
	    search {
		if { $search_state eq "any" } {
		    $self manage_database search_any
		}
		if { $search_state } {
		    set _clear_save_entries_recursive_dict ""
		    foreach w $tabs {
		        $self _clear_save_entries_recursive $w
		    }
		    bind notebook_children <KeyPress> [mymethod manage_database update_idle]
		    bind notebook_children <Escape> [mymethod manage_database search]
		    bind notebook_children <Return> [mymethod manage_database fill_entries]
		} else {
		    foreach w $tabs {
		        $self _restore_entries_recursive $w
		    }
		    bind notebook_children <KeyPress> ""
		    bind notebook_children <Escape> ""
		    bind notebook_children <Return> ""
		    $self manage_database update
		}
	    }
	    update_idle {
		if { [info exist db_update_idle] } { after cancel $db_update_idle }
		set db_update_idle [after 100 [list catch [mymethod manage_database update]]]
	    }
	    update {
		set optional {
		    { -see item "" }
		}
		set compulsory ""
		parse_args $optional $compulsory $args

		if { [info exist db_update_idle] } {
		    after cancel $db_update_idle
		    unset db_update_idle
		}
		set sel_ids ""
		foreach item [$win.list selection get] {
		    lappend sel_ids [$win.list item text $item 0]
		}
		set sel_ids [lsort -integer $sel_ids]
		set sort_column [$win.list sort_column]
		set yview [$win.list yview]
		$win.list item delete all

		if { ![$db sql table_exists $table] } {
		    return
		}
		set cols [$db sql column_names $table]

		set selcols [lrange $cols 0 2]
		foreach col [lrange $cols 3 end] {
		    if { ![regexp {(.*)__(.*)} $col {} name subname] } {
		        lappend selcols $col
		    }
		}
		set selcols_p "\"[join $selcols \",\"]\""

		if { $search_state eq "any" } {
		    set v [string trim [$win.f.db.searchframe.e get]]
		    set cndList ""
		    if { $v ne "" } {
		        set vp [$db sql escape $v]
		        foreach col [concat id [lrange $cols 3 end]] {
		            if { [regexp {__} $col] } { continue }
		            set pat '%[string map [list \\ \\\\ % \\% _ \\_] \
		                    $vp]%'
		            lappend cndList "\"$col\" like $pat"
		        }
		    }
		    set cmd "select $selcols_p from $table"
		    if { [llength $cndList] } {
		        append cmd " where " [join $cndList " or "]
		    }
		} elseif { $search_state } {
		    set cndList ""
		    foreach i [dict keys $_clear_save_entries_recursive_dict] {
		        set err [catch { $i cget -textvariable } var]
		        if { !$err && [regexp {formulae::uvalues\(.*,(.*)\)} $var {} fieldname] } {
		            set ipos [lsearch -exact $cols $fieldname]
		            if { $ipos != -1 } {
		                set v [string trim [$i get]]
		            } else { set err 1 }
		        } else { set err 1 }
		        if { $err } {
		            set err [catch { $i cget -clientdata } fieldname]
		            if { !$err } {
		                set ipos [lsearch -exact $cols $fieldname]
		                if { $ipos != -1 } {
		                    set v [$i get 1.0 end-1c]
		                } else { set err 1 }
		            }
		        }
		        if { $err } { continue }
		        set v [string trim $v]
		        if { $v eq "" } { continue }
		        set vp [$db sql escape $v]
		        set pat '%[string map [list \\ \\\\ % \\% _ \\_] \
		                $vp]%'
		        lappend cndList "\"$fieldname\" like $pat"
		    }
		    set cmd "select $selcols_p from \"$table\""
		    if { [llength $cndList] } {
		        append cmd " where " [join $cndList " and "]
		    }
		} else {
		    set cmd "select $selcols_p from \"$table\""
		}
		if { [info exists ::ispocket] && $::ispocket } {
		    append cmd " limit 20"
		}
		$db sql maplist $cmd {
		    set v [list $id]
		    foreach i [lrange $selcols 3 end] {
		        lappend v [set $i]
		    }
		    $win.list insert end $v
		    if { [lsearch -sorted -integer $sel_ids $id] != -1 } {
		        $win.list selection add end
		    }
		}
		eval [list $win.list sort_column] $sort_column
		$win.list yview moveto [lindex $yview 0]
		if { $see ne "" } {
		    $win.list see $see
		}
	    }
	    contextual {
		foreach "- - x y" $args break
		destroy $win._cmenu
		set menu [menu $win._cmenu -tearoff 0]
		
		$menu add command -label [_ "Edit"] -command [mymethod \
		        manage_database fill_entries]
		$menu add command -label [_ "Search"] -command [mymethod \
		        manage_database search_any]
		$menu add command -label [_ "SQL table schema"] -command \
		    [mymethod manage_database output_schema]
		$menu add separator
		$menu add command -label [_ "Delete"] -command \
		    [mymethod manage_database delete]
		
		tk_popup $menu $x $y
	    }
	    output_schema {
		set table_schema [$db sql table_schema $table]

		destroy $win._output_schema
		set w [dialogwin_snit $win._output_schema -title \
		        [_ "SQL table schema"] -okname -]
		set f [$w giveframe]
		wordwidget_snit $f.t -width 45 -height 15  -readonly 1
		catch { $f.t configure -bg [$f cget -bg] }
		pack $f.t -fill both -expand 1
		
		set xml "<lognoter>"
		foreach line [split $table_schema \n] {
		    append xml "<para>[xml_map $line]</para>\n"
		}
		append xml "</lognoter>"
		#$f.t ApplyXMLdataAsText $xml
		after 100 [list $f.t ApplyXMLdataAsText $xml]
		set action [$w createwindow]
		destroy $w
	    }
	}
    }
}

proc formulae::create_window { wp xmldata } {

    dom parse $xmldata doc
    $doc documentElement root

    return [create_windowD $wp $root ""]
}

proc formulae::create_windowD { args } {
    variable uvalues
    variable uvalues_moredata
    variable update_stack
    variable svg_stack
    variable nicer_name
    variable update_callback
    variable window_doc
    variable current_set_name
    variable current_set_print
    variable window_parent
    variable state
    variable tabstyle
    variable lognoter_db
    variable database_default_table
    variable page_tcl_code

    set optional {
	{ -tabstyle arrows|arrows_small|notebook arrows }
	{ -update_callback callback "" }
	{ -state normal|disabled normal }
	{ -lognoter_db db "" }
	{ -database_default_table table "" }
	{ -maintain_grid boolean 1 }
    }
    set compulsory "wp root"
    parse_args $optional $compulsory $args

    if { $root ne "" } {
	set page_tcl_code ""
	foreach node [$root selectNodes //tcl] {
	    append page_tcl_code "[$node text]\n"
	}
	set window_doc [dom createDocument [$root nodeName]]
	set new_root [$window_doc documentElement]
	foreach att [$root attributes] {
	    $new_root setAttribute $att [$root @$att]
	}
	foreach node [$root childNodes] {
	    $new_root appendChild $node
	}
	if { [$new_root @tabstyle ""] ne "" } {
	    set tabstyle [$new_root @tabstyle]
	}
    } else {
	unset -nocomplain window_doc
    }

    set window_parent $wp
    set current_set_name ""
    set update_stack ""
    set svg_stack ""
    array unset uvalues
    array unset uvalues_moredata
    array unset nicer_name

    package require tile
    #package require BWidget
    package require wordwidget_snit
    package require tooltip
    namespace eval :: { namespace import -force tooltip::tooltip }


    create_images

    if { $root eq "" } {
	#set update_callback ""
	edit_header -update_callback $update_callback $wp
	set ret ""
    } else {
	set ret [create_windowD_do -maintain_grid $maintain_grid]
    }
    return $ret
}

# create the formulae window
# args can be: ?-tabstyle arrows|arrows_small|notebook? ?-showhidden boolean?
proc formulae::create_windowD_do { args } {
    variable uvalues
    variable uvalues_moredata
    variable update_stack
    variable svg_stack
    variable nicer_name
    variable window_doc
    variable current_set_name
    variable current_set_print
    variable current_set_print_all_none
    variable window_parent
    variable lognoter_db
    variable state
    # can be arrows or arrows_small or notebook
    variable tabstyle
    variable notebook_widget
    variable page_tcl_code

    set optional {
	{ -tabstyle tabstyle arrows }
	{ -showhidden boolean 0 }
	{ -maintain_grid boolean 1 }
    }
    set compulsory ""
    parse_args $optional $compulsory $args

    if { [info exists window_doc] } {
	set root [$window_doc documentElement]
	if { [$root @tabstyle ""] ne "" } {
	    set tabstyle [$root @tabstyle]
	}
    } else { set root "" }

    set wp $window_parent

    set current_set_name ""
    set update_stack ""
    set svg_stack ""
    array unset uvalues
    array unset uvalues_moredata
    array unset nicer_name

    if { [interp exists formulae] } { interp delete formulae }
    interp create formulae

    formulae eval [list set auto_path $::auto_path]
    #interp alias formulae ::tcl::mathfunc::defaultvalue "" formulae::defaultvalue $root
    interp alias formulae ::tcl::mathfunc::max "" formulae::mathfunc_max
    interp alias formulae ::tcl::mathfunc::min "" formulae::mathfunc_min
    interp alias formulae ::tcl::mathfunc::case "" formulae::mathfunc_case
    interp alias formulae ::tcl::mathfunc::pi "" formulae::mathfunc_pi
    interp alias formulae ::tcl::mathfunc::interpolate "" formulae::mathfunc_interpolate
    interp alias formulae ::tcl::mathfunc::dictionary "" formulae::mathfunc_dictionary
    interp alias formulae xml "" formulae::xml
    if { $lognoter_db ne "" } {
	interp alias formulae lognoter_db "" $lognoter_db
	interp alias formulae executefile "" $lognoter_db executefile
	interp alias formulae sql "" $lognoter_db sql
	interp alias formulae refresh "" $wp.n manage_database update
    }
    set len [expr {[winfo width $wp]-5}]

    if { $page_tcl_code ne "" } {
	set err [catch { formulae eval $page_tcl_code } errstring]
	if { $err } {
	    snit_messageBox -message $errstring
	} 
    }

    set active_tab 0
    if { [winfo exists $wp.n] } {
	catch { $wp.n index current } active_tab
	bind $wp.n <Destroy> ""
	if { $maintain_grid } {
	    set grid_info [grid info $wp.n]
	    if { $grid_info eq "" } {
		set grid_info [winfo manager $wp.n]
	    }
	}
	destroy $wp.n
    } else { set active_tab 0 }

    set usableviewmode [$root @usableviewmode 0]
    set updatebutton [$root @updatebutton 1]

    if { $usableviewmode } {
	set db_state normal
    } else {
	set db_state $state
    }

    snit_notebook notebook_type $tabstyle
    snit_notebook $wp.n -lognoter_db $lognoter_db \
	-database [$root @database ""] -rootnode $root \
	-dbstate $db_state
    set notebook_widget $wp.n

    if { [info exists grid_info] } {
	if { $grid_info eq "paned" } {
	    set p [lindex [grid slaves $wp] 0]
	    $p insert 0 $wp.n -weight 1
	} elseif { $grid_info ne "" } {
	    eval grid $wp.n $grid_info
	}
    }

    bind $wp.n <Destroy> [namespace code [list destroy_window]]
    bind [winfo toplevel $wp] <Control-h> [namespace code [list create_windowD_do \
		-showhidden [expr {($showhidden)?0:1}]]]

    if { $root eq "" } { return $wp.n }

    foreach "idx idxw numlf" [list 0 0 0] break
    set f [frame $wp.n.f[incr idx]]
    grid columnconfigure $f 0 -weight 1

    $wp.n add $f -sticky nsew -text [_ "Intro"] -image [cu::get_icon iconthemes-16] \
	-compound left

    if { $state eq "normal" } {
	bind $f <ButtonRelease-3> [list formulae::header_contextual_menu $wp \
		%X %Y]
    }
    foreach node [$root selectNodes title|description|container|setname] {
	switch [$node nodeName] {
	    title {
		set l [ttk::label $f.l[incr idxw] -text [$node text] \
		        -font "-weight bold" -wraplength [expr {$len-16}] -justify left]
		if { $updatebutton } {
		    set b [ttk::button $f.b[incr idxw] -image [cu::get_icon iconthemes-16]]
		    $b configure -command [list formulae::window_actualize \
		            -conditions_window $b]
		    if { $state eq "disabled" && !$usableviewmode } {
		        $b state disabled
		    }
		    grid $l $b -sticky nw -padx 2
		    tooltip $b [_ "Update values (Return)"]
		} else {
		    grid $l -sticky nw -padx 2
		}
		if { $state eq "normal" } {
		    bind $l <ButtonRelease-3> [list formulae::header_contextual_menu $wp \
		            %X %Y]
		}
	    }
	    description {
		set txt [string trim [$node text]]
		if { $txt ne "" } {
		    set l [ttk::label $f.l[incr idxw] -text $txt \
		        -wraplength $len -justify left]
		    grid $l -sticky nw -columnspan 2 -pady 2 -padx 2
		    if { $state eq "normal" } {
		        bind $l <ButtonRelease-3> [list formulae::header_contextual_menu $wp \
		                %X %Y]
		    }
		}
	    }
	    setname {
		set values ""
		foreach "numprinted numtotal" [list 0 0] break
		foreach cnode [$node selectNodes name] {
		    lappend values [$cnode @n]
		    incr numprinted [$cnode @print 1]
		    incr numtotal
		}
		set current_set_name [$node @n ""]
		set current_set_print [$node @print 1]
		if { $numprinted == $numtotal } {
		    set current_set_print_all_none all
		} elseif { $numprinted > 0 } {
		    set current_set_print_all_none some
		} else {
		    set current_set_print_all_none none
		}

		set f1 [frame $f.f[incr idxw]]
		set l [ttk::label $f1.l[incr idxw] -text [_ "Load case:"]]
		set cb [ttk::combobox $f1.cb[incr idxw] -textvariable \
		        formulae::current_set_name -values $values]
		if { $state eq "disabled" } {
		    if { $usableviewmode } {
		        $cb state readonly
		    } else {
		        $cb state disabled
		    }
		}
		focus $cb
		bindtags $cb [concat formulae [bindtags $cb]]
		bind $cb <<ComboboxSelected>> [list formulae::manage_names select \
		        $wp $f1.cb$idxw $node]
		bind $cb <Return> [list formulae::manage_names accept $wp $cb $node]
		set mb [ttk::menubutton $f1.mb[incr idxw] -image [cu::get_icon bookmark_add-16] -style \
		    Toolbutton -menu $f1.mb$idxw.m -state $state]
		menu $f1.mb$idxw.m -tearoff 0
		$f1.mb$idxw.m add command -label [_ "Accept"] -command \
		    [list formulae::manage_names accept $wp $cb $node] \
		    -acc "Return"
		$f1.mb$idxw.m add separator
		$f1.mb$idxw.m add checkbutton -label [_ "Print"] -variable \
		    formulae::current_set_print -command \
		    [list formulae::manage_names print $wp $cb $node]
		$f1.mb$idxw.m add separator
		$f1.mb$idxw.m add radiobutton -label [_ "Print all"] -command \
		    [list formulae::manage_names printall $wp $cb $node] -variable \
		    formulae::current_set_print_all_none -value all
		$f1.mb$idxw.m add radiobutton -label [_ "Print none"] -command \
		    [list formulae::manage_names printnone $wp $cb $node] -variable \
		    formulae::current_set_print_all_none -value none
		$f1.mb$idxw.m add separator
		$f1.mb$idxw.m add command -label [_ "Delete"] -command \
		    [list formulae::manage_names delete $wp $cb $node]
	    
		grid $l $cb $mb -sticky nwe
		grid columnconfigure $f 1 -weight 1
		grid $f1 -sticky nw -columnspan 2 -pady 2 -padx 2
	     
		if { $state eq "normal" } {
		    foreach i [list $l $cb $mb] {
		        bind $i <ButtonRelease-3> [list formulae::header_contextual_menu $wp \
		                %X %Y]
		    }
		}   
	    }
	    container {
		if { [$node @location ""] ne "same_page" } {
		    if { $numlf != 0 } { incr idx }
		    set txt ""
		    foreach i [list pns pn n] {
		        if { [$node @$i ""] ne "" } {
		            set txt [$node @$i]
		            break
		        }
		    }
		    if { $txt eq "" } {
		        set txt [_ "Step %d" $idx]
		    }
		    if { $numlf != 0 } {
		        set f [frame $wp.n.f$idx]
		        grid columnconfigure $f 0 -weight 1
		        grid rowconfigure $f 1 -weight 1
		        $wp.n add $f -sticky nsew -text $txt -image \
		            [cu::get_icon contents2-16] -compound left
		        foreach "idxw numlf" [list 1 0 0] break
		    } else {
		        $wp.n tab 0 -text $txt
		    }
		}
		set cstate [$node @state normal]
		if { $cstate eq "hidden" && $showhidden } { set cstate normal }
		if { $cstate eq "hidden" } {
		    $wp.n tab $f -state hidden
		} elseif { [$node @condition ""] ne "" } {
		    set cmd formulae::check_container_condition
		    lappend cmd [$node @n ""] [$node @condition ""] $wp.n $f
		    lappend update_stack $cmd
		}
		if { [$node @help ""] ne "" } {
		    set l [ttk::label $f.l[incr idxw] -text [string trim [$node @help]] \
		            -wraplength [expr {$len-16}] -justify left]
		    set b [ttk::button $f.b[incr idxw] -image \
		            [cu::get_icon iconthemes-16] \
		            -command formulae::window_actualize]
		    grid $l $b -sticky nw -padx 2 -pady 2
		    tooltip $b [_ "Update values (Return)"]
		    if { $state eq "normal" } {
		        bind $l <ButtonRelease-3> [list formulae::entry_contextual_menu $wp \
		                [$node @n] container %X %Y]
		    }
		}
		set l [ttk::labelframe $f.l[incr idxw] -text [$node @pn [$node @n]]]
		grid $l -sticky nsew -columnspan 2 -padx 5 -pady "2 5"
		set ncols [lindex [grid size $f] 1]
		foreach i [range 0 $ncols] { grid rowconfigure $f $i -weight 0 }
		grid rowconfigure $f $ncols -weight 1

		set ns { svg http://www.w3.org/2000/svg }
		set svgNode [$node selectNodes -namespaces $ns svg:svg]
		if { $svgNode ne "" } {
		    package require xml2pdf
		    ttk::paned $l.p -orient horizontal
		    $l.p add [frame $l.p.f1]
		    canvas $l.p.c -bg white
		    set svg2draw [svg2draw %AUTO% -destroy_with_canvas 1 -canvas $l.p.c]
		    lappend svg_stack $svg2draw
		    if { $state eq "normal" } {
		        $svg2draw configure -add_to_contextual \
		            [list formulae::add_drawing_contextual_menu [$node @n]]
		    }
		    catch { $svg2draw draw -svgNode $svgNode }
		    $l.p add $l.p.c
		    grid $l.p -sticky nwes
		    grid columnconfigure $l 0 -weight 1
		    grid rowconfigure $l 0 -weight 1
		    set l $l.p.f1
		}
		process_window_container $wp $l $node $state $usableviewmode $showhidden
		if { $state eq "normal" } {
		    bind $l <ButtonRelease-3> [list formulae::entry_contextual_menu $wp \
		            [$node @n] container %X %Y]
		}

		if { ![llength [winfo children $l]] } {
		    $l configure -height 50
		}
		
		if { $state eq "normal" } {
		    bind $l <ButtonRelease-3> [list formulae::entry_contextual_menu $wp \
		            [$node @n] container %X %Y]
		}
		incr numlf
	    }
	}
    }
    window_actualize
    catch { $wp.n select $active_tab }
    $wp.n _compute_size

    if { [winfo manager $wp.n] eq "paned" } {
	set p [lindex [grid slaves $wp] 0]
	update
	set h [winfo reqheight $wp.n]
	$p sashpos 0 $h
    }
    return $wp.n
}

# create window for container
proc formulae::process_window_container { key w containerNode state usableviewmode \
    showhidden } {

    variable uvalues
    variable update_stack
    variable nicer_name

    set wp $w
    while { $wp ne "." && [winfo width $wp] < 2 } { set wp [winfo parent $wp] }
    set len [expr {[winfo width $wp]-5}]
    
    set idxw 0
    foreach node [$containerNode selectNodes param|condition] {
	if { [$node nodeName] eq "param" } {
	    set nicer_name([$node @n ""]) [$node @pn [$node @n]]
	}
	set node_state [$node @state normal]
	if { $showhidden } { set node_state normal }

	if { [string index [$node @value ""] 0] == "\[" } {
	    set is_formula 1
	    if { [$node @field_type ""] eq "formatted expression" || \
		[$node @field_type ""] eq "text expression" } {
		set is_formula 0
	    }
	} else {
#             set err [catch { expr {[$node @value]*1.0} } v]
#             if { $err } { set v "!ERROR" }
	    set v [$node @value ""]
	    formulae eval [list set [$node @n] $v]
	    set is_formula 0
	}
	set ws ""
	if { !$is_formula && $node_state eq "hidden" } {
	    #nothing
	} elseif { $is_formula && $node_state eq "hidden" } {
	    set cmd [list calculate_and_update_formula [$node nodeName]]
	    foreach i [list n pn value units true false] {
		lappend cmd [$node @$i ""]
	    }
	    lappend cmd ""
	    eval $cmd
	    lappend update_stack $cmd
	} elseif { !$is_formula } {
	    set values [split [$node @values ""] ,]
	    if { [$node @dict ""] ne "" } {
		set dict [split [$node @dict] ,]
		set values_trans ""
		foreach v_in $values {
		    if { [dict exists $dict $v_in] } {
		        lappend values_trans [dict get $dict $v_in]
		    } else {
		        lappend values_trans $v_in
		    }
		}
	    } else {
		set dict ""
		set values_trans $values
	    }
	    if { [llength $values] == 2 && \
		[string is boolean -strict [lindex $values 0]] && \
		    [string is boolean -strict [lindex $values 1]] } {
		set isboolean 1
	    } else { set isboolean 0}

	    foreach "gridList wList" [list "" ""] break

	    if { $isboolean || [$node @command ""] ne "" } {
		set l ""
	    } elseif { [regexp {<su(p|b)} [$node @pn [$node @n]]] } {
		set l [wordwidgetformula $w.wf[incr idxw] \
		        -maxwidth 100 -formula [$node @pn [$node @n]]:]
		lappend gridList $l
		lappend wList $l
	    } else {
		set l [ttk::label $w.l[incr idxw] -text [$node @pn [$node @n]]: \
		        -takefocus 0]
		lappend gridList $l
		lappend wList $l
	    }
	    switch -- [$node @units ""] "" { set isnumber 0 } default { set isnumber 1 }

	    set sticky nw
	    set field_type [$node @field_type ""]
	    if { $field_type eq "date" } {
		set e [cu::dater_entry $w.e[incr idxw] -textvariable \
		        formulae::uvalues($key,[$node @n])]
		lappend gridList $e
		lappend wList $e
	    } elseif { $field_type eq "file" } {
		set f [ttk::frame $w.f[incr idxw]]
		set e [ttk::entry $f.e -textvariable \
		        formulae::uvalues($key,[$node @n])]
		set b [cu::menubutton_button $f.b \
		        -image [cu::get_icon filenew-16] -style Toolbutton]
		grid $e $b -sticky w
		$b configure -menu $b.m -command [list formulae::fieldtype_file_ops \
		        $w "$e $b" open_browser $key $node]
		menu $b.m -tearoff 0
		$b.m add command -image [cu::get_icon filenew-16] -label [_ "Select file"] \
		    -command [list formulae::fieldtype_file_ops \
		        $w "$e $b" open_browser $key $node] -compound left
		$b.m add command -image [cu::get_icon remove-16] -label [_ "Clear"] \
		    -command [list formulae::fieldtype_file_ops \
		        $w "$e $b" clear $key $node] -compound left
		$b.m add separator
		$b.m add command -label [_ "Execute"] \
		    -command [list formulae::fieldtype_file_ops \
		        $w "$e $b" execute $key $node]
		$b.m add separator
		if { $::tcl_platform(platform) == "windows" } {
		    set dirs ""
		    set rkey {HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders}
		    foreach i [list Personal Desktop] {
		        lappend dirs [registry get $rkey $i]
		    }
		    lappend dirs $::env(TEMP)
		    set tmpdir $::env(TEMP)
		} else {
		    set dirs [list $::env(HOME) /tmp]
		    set tmpdir /tmp
		}
		foreach dir $dirs {
		    regsub {([\\/]).*([\\/])} $dir {\1...\2} dir_nice
		    $b.m add command -label [_ "Extract in '%s'" $dir_nice] \
		        -command [list formulae::fieldtype_file_ops \
		            $w "$e $b" extract $key $node $dir]
		}
		$b.m add command -label [_ "Extract in ..."] \
		    -command [list formulae::fieldtype_file_ops \
		        $w "$e $b" extract $key $node ""]
		$b.m add separator
		$b.m add command -label [_ "RamViewer"] \
		    -command [list formulae::fieldtype_file_ops \
		        $w "$e $b" ramviewer $key $node]
		if { [info command dnd] ne "" } {
		    foreach i [list $e $b] {
		        dnd bindtarget $i text/uri-list <Drop> [list formulae::fieldtype_file_ops \
		                $w "$e $b" open_files $key $node %D]
		    }
		}
		bind $e <<Paste>> [list formulae::fieldtype_file_ops \
		        $w "$e $b" paste $key $node $e]
		lappend gridList $f
		lappend wList $e $b

		set cmd [list formulae::fieldtype_file_ops $w $wList update_tooltip \
		        $key $node]
		trace add variable formulae::uvalues($key,[$node @n]) write "$cmd;#"
		bind $e <Destroy> [list trace remove variable \
		        formulae::uvalues($key,[$node @n]) write "$cmd;#"]
	    } elseif { [lsearch [list "long text" "formatted" "formatted expression" \
		"text expression"] \
		$field_type] != -1 } {
		set e [wordwidget_and_toolbox $w.e[incr idxw] \
		        -highlightthickness 1 -bd 0 -clientdata [$node @n]]
		if { [lsearch [list "long text" "text expression"] \
		    $field_type] != -1 } {
		    $e configure -mode off
		}
		set sticky we
		lappend gridList $e
		lappend wList $e
		if { [lsearch [list "formatted expression" "text expression"] \
		    $field_type] != -1 } {
		    set cmd [list formulae::calculate_and_update_formatted $e \
		            value [$node @field_type] $key [$node @n] [$node @value]]
		    lappend update_stack $cmd
		    $e configure -readonly 1 -bd 0
		    eval $cmd
		} else {
		    set cmd [list formulae::update_variable_formatted $e \
		            $field_type $key [$node @n]]
		    trace add variable formulae::uvalues($key,[$node @n]) read \
		        "$cmd;#"
		    bind $e <Destroy> [list trace remove variable \
		            formulae::uvalues($key,[$node @n]) read "$cmd;#"]
		}
		set cmd [list formulae::calculate_and_update_formatted $e \
		        variable [$node @field_type] $key [$node @n] [$node @value]]
		trace add variable formulae::uvalues($key,[$node @n]) write \
		    "$cmd;#"
		bind $e <Destroy> [list trace remove variable \
		        formulae::uvalues($key,[$node @n]) write "$cmd;#"]
	    } elseif { $isboolean } {
		set e [ttk::checkbutton $w.e[incr idxw] -text [$node @pn [$node @n]] \
		        -variable formulae::uvalues($key,[$node @n]) \
		        -command [list formulae::entry_validate %W [$node @n] \
		            $key $isnumber]]
		lappend gridList $e
		lappend wList $e
	    } elseif { [$node @command ""] ne "" } {
		set cmd [list formulae eval [$node @command]]
		set e [ttk::button $w.e[incr idxw] -text [$node @pn [$node @n]] \
		        -image [$node @img ""] -command $cmd]
		if { [$node @pn ""] ne "" && [$node @img ""] ne "" } {
		    $w.e$idxw configure -compound left
		}
		lappend gridList $e
		lappend wList $e
	    } elseif { [llength $values] == 2 && [$node @editable] == 0 } {
		ttk::frame $w.e[incr idxw]
		ttk::radiobutton $w.e$idxw.r1 -text [lindex $values_trans 0] \
		    -variable formulae::uvalues($key,[$node @n]) \
		    -value [lindex $values_trans 0]\
		    -command [list formulae::entry_validate %W [$node @n] \
		        $key $isnumber]
		ttk::radiobutton $w.e$idxw.r2 -text [lindex $values_trans 1] \
		    -variable formulae::uvalues($key,[$node @n]) \
		    -value [lindex $values_trans 1]\
		    -command [list formulae::entry_validate %W [$node @n] \
		        $key $isnumber]
		grid $w.e$idxw.r1 $w.e$idxw.r2 -sticky w -padx 2
		set e [list $w.e$idxw.r1 $w.e$idxw.r2]
		lappend gridList $w.e$idxw
		lappend wList $w.e$idxw.r1 $w.e$idxw.r2
	    } elseif { [llength $values] == 0 } {
		set e [ttk::entry $w.e[incr idxw] -width 20 -textvariable \
		        formulae::uvalues($key,[$node @n]) \
		        -validate focusout -validatecommand \
		        [list formulae::entry_validate %W [$node @n] $key $isnumber] \
		        -invalidcommand [list formulae::entry_invalid %W]]
		lappend gridList $e
		lappend wList $e
	    } else {
		set e [ttk::combobox $w.e[incr idxw] -width 17 -textvariable \
		        formulae::uvalues($key,[$node @n]) \
		        -validate focusout -validatecommand \
		        [list formulae::entry_validate %W [$node @n] $key $isnumber] \
		        -invalidcommand [list formulae::entry_invalid %W] \
		        -values $values_trans]
		if { [$node @editable] == 0 } {
		    $e state readonly
		    set cmd [list formulae::update_non_editable_field $e \
		            $key [$node @n]]
		    trace add variable formulae::uvalues($key,[$node @n]) write \
		        "$cmd;#"
		    bind $e <Destroy> [list trace remove variable \
		            formulae::uvalues($key,[$node @n]) write "$cmd;#"]
		}
		lappend gridList $e
		lappend wList $e
	    }
	    if { $state eq "disabled" && !$usableviewmode } {
		foreach e_in $e { $e_in state disabled }
	    }
	    foreach e_in $e {
		bindtags $e_in [concat formulae [bindtags $e_in]]
		bind $e_in <Return> [list formulae::window_actualize_entry %W \
		        [$node @n] $key $isnumber]
	    }
	    if { $field_type ne "formatted expression" && \
		$field_type ne "text expression"} {
		if { $dict eq "" || ![dict exists $dict $v] } {
		    set uvalues($key,[$node @n]) $v
		} else {
		    set uvalues($key,[$node @n]) [dict get $dict $v]
		}
	    } else {
	       #set uvalues($key,[$node @n]) ""
	    }
	    if { [$node @units ""] ne "" } {
		set u [ttk::label $w.l[incr idxw] -text [$node @units ""]]
		lappend gridList $u
		lappend wList $u
	    } else { set u "" }

	    eval grid $gridList -sticky $sticky -pady 1
	    if { $l ne "" } { grid configure $l -padx "2 0" }
	    if { $u ne "" } { grid configure $u -padx 2 }
	    if { [regexp {e} $sticky] && [regexp {w} $sticky] } {
		grid configure [lindex $gridList end] -padx "0 2"
		grid columnconfigure $w [expr {[llength $gridList]-1}] -weight 1
	    } else {
		grid columnconfigure $w [llength $gridList] -weight 1
	    }
	    eval lappend ws $wList
	} else {
	    set f [frame $w.f[incr idxw]]
	    set fp [winfo parent $f]
	    while 1 {
		set err [catch { $fp cget -font } font]
		if { $err } {
		    if { $fp eq "." } {
		        set font TkDefaultFont
		        break
		    }
		} else { break }
		set fp [winfo parent $fp]
	    }
	    set fontsize [font actual $font -size]
	    set wimg [wordwidgetformula $f.wf1 -font $font -normalize_spaces 0\
		    -fontsize $fontsize -maxwidth [expr {int(1.0*$len)}]]
	    grid $wimg -sticky w
	    grid $f -sticky nw -columnspan 3 -padx "2 0" -pady 2
	    lappend ws $wimg

	    set cmd [list calculate_and_update_formula [$node nodeName]]
	    foreach i [list n pn value units true false] {
		lappend cmd [$node @$i ""]
	    }
	    lappend cmd $wimg
	    eval $cmd
	    lappend update_stack $cmd
	    
	    set wList $wimg
	}
	if { $state eq "normal" } {
	    foreach i $wList {
		bind $i <ButtonRelease-3> [list formulae::entry_contextual_menu $wp \
		        [$node @n] param %X %Y]
	    }
	}
	if { [$node hasAttribute help] } {
	    foreach i $ws {
		tooltip $i [$node @help]
	    }
	}
    }
}

proc formulae::check_container_condition { name condition nb tab } {
    if { $condition ne "" } {
	if { [string index $condition 0] == "\[" } {
	    set value [string range $condition 1 end-1]
	    set err [catch { eval $value } condition]
	    if { $err } { set  condition 1 }
	} else {
	    error [_ "error in container '%s'. Condition not ok" $name]
	}
    } else { set condition 1 }

    if { !$condition } {
	$nb tab $tab -state disabled
    } else {
	$nb tab $tab -state normal
    }
}

proc formulae::update_variable_formatted { wp field_type key name } {
    variable uvalues

    switch $field_type {
	"long text" {
	    set uvalues($key,$name) [string trim [$wp get 1.0 end-1c]]
	}
	default {
	    set uvalues($key,$name) "<wordwidget>[$wp exportXML]</wordwidget>"
	}
    }
}
proc formulae::calculate_and_update_formatted { wordwidget type field_type key name value } {
    variable uvalues

    if { $type eq "variable" } {
	set v $uvalues($key,$name)
	if { [regexp {^<wordwidget>|<lognoter>} $v] } {
	    set vp $v
	} else {
	    set vp "<wordwidget>"
	    foreach line [split $v \n] {
		append vp "<para>[xml $line]</para>"
	    }
	    append vp "</wordwidget>"
	}
    } elseif { [string index $value 0] eq "\[" } {
	set value [string range $value 1 end-1]
	set err [catch { formulae eval $value } v]
	if { $err } {
	    set vp "<wordwidget><para>[xml "error: $v"]</para></wordwidget>"
	    set v "error: $v"
	} elseif { $field_type eq "text expression" } {
	    set _ "<wordwidget>"
	    set numlines 0
	    foreach line [split $v \n] {
		append _ "<para>[xml $line]</para>"
		incr numlines
	    }
	    append _ "</wordwidget>"
	    append vp $_
	    if { $numlines > 4 } { set numlines 4 }
	    $wordwidget configure -height $numlines
	} else {
	    set vp $v
	    if { [string trim $vp] eq "" } { set vp "<wordwidget/>" }
	}
    } else {
	set v $value
	if { [regexp {^<wordwidget>|<lognoter>} $v] } {
	    set vp $v
	} else {
	    set vp "<wordwidget>"
	    foreach line [split $v \n] {
		append vp "<para>[xml $line]</para>"
	    }
	    append vp "</wordwidget>"
	}
    }
    $wordwidget ClearText
    set err [catch { $wordwidget ApplyXMLdataAsText $vp } errstring]
    if { $err } {
	set vp "<wordwidget><para>[xml "error: $errstring"]</para></wordwidget>"
	$wordwidget ApplyXMLdataAsText $vp
    }
    formulae eval [list set $name $v]
}

# calculate_and_update_formula
proc formulae::calculate_and_update_formula { ntype name pname value units true false wimg } {

    set errorMessage ""
    set value [string range $value 1 end-1]
    set valueE [string map [list \[ \\\[] $value]
    set err [catch { eval $valueE } v]
    if { $err } {
	set errorMessage $v
	set v !ERROR
    }
    
    if { $pname ne "" } { set txt "$pname=" } else { set txt "" }   
    if { [string match "e *" $value] } {
	set dosimple 1
	if { $ntype eq "condition" && [regexp {[<>]} $value] } {
	    #set f [string range $value 2 end]
	    set f [lindex $value 1]
	    regexp {(.*)([<>]=?)(.*)} $f {} v1 sign v2
	    set v1_e [string map [list \[ \\\[] $v1]
	    set v2_e [string map [list \[ \\\[] $v2]
	    set err 0
	    if { !$err } { set err [catch { eval "e $v1_e" } v1_res] }
	    if { !$err } { set err [catch { eval "e $v2_e" } v2_res] }
	    if { !$err } { set err [catch { nicer_formula $v1 } v1_f] }
	    if { !$err } { set err [catch { nicer_formula $v2 } v2_f] }
	    if { !$err && [isdouble $v1_res] && [isdouble_or_void $v2_res]} {
		if { ![string is double -strict $v1_f] } {
		    append txt "$v1_f=[format %.4g $v1_res] "
		} else {
		    append txt "$v1_f "
		}
		append txt "$sign "
		if { ![string is double -strict $v2_f] } {
		    append txt "$v2_f=[format %.4g $v2_res]"
		} else {
		    append txt "$v2_f"
		}
		set dosimple 0
	    }
	}
	if { $dosimple } {
	    #set f [string range $value 2 end]
	    set f [lindex $value 1]
	    set err [catch { nicer_formula $f } ret]
	    if { !$err } {
		append txt $ret
	    } else {
		set errorMessage $ret
		append txt [_ "!ERROR"]
	    }
	}
    }
    if { $ntype eq "param" } {
	formulae eval [list set $name $v]
	if { $v ne "!ERROR" } {
	    set num [format %.4g $v]
	} else { set num "!ERROR" }
	if { [regexp {(.*)e([+-])(\d+)} $num {} p1 s exp] } {
	    if { $s eq "+" } { set s "" }
	    if { $exp != 0 } { set exp [string trimleft $exp 0] }
	    set num "$p110<sup>$s$exp</sup>"
	}
	append txt "=$num"
	if { $units ne "" } {
	    append txt " [nicer_units $units]"
	}
    } else {
	if { $units ne "" } {
	    append txt " [nicer_units $units]"
	}
	if { $v eq "!ERROR" } {
	    append txt "     [_ "!ERROR"]"
	} elseif { $v } {
	    if { $true eq "" } { set true "[_ OK#C#Valid]" }
	    append txt "     $true (ok)"
	} else {
	    if { $false eq "" } { set false "[_ {Fail#C#No Valid}]" }
	    append txt "     $false (fail)"
	}
    }
    if { $wimg ne "" } {
	$wimg configure -formula $txt
	if { $errorMessage ne "" } {
	    tooltip $wimg $errorMessage
	}
    }
}

proc formulae::entry_validate { entry name key isnumber } {
    variable uvalues

    if { $isnumber } {
	set ret [isdouble_or_void $uvalues($key,$name)]
    } else { set ret 1 }
    if { $ret } {
	formulae eval [list set $name $uvalues($key,$name)]
    }
    return $ret
}

proc formulae::entry_invalid { entry } {
    $entry selection range 0 end
    focus $entry
    bell
}

proc formulae::update_non_editable_field { e key n } {
    variable uvalues

    set values [$e cget -values]
    set v $uvalues($key,$n)
    if { [lsearch -exact $values $v] == -1 } {
	set uvalues($key,$n) [lindex $values 0]
    }
}

proc formulae::fieldtype_file_ops { w wList what key node args } {
    variable lognoter_db
    variable uvalues
    variable uvalues_moredata

    switch $what {
	open_browser {
	    if { $lognoter_db ne "" } {
		set lastdir [$lognoter_db getpreference lastdir]
	    } else {
		set lastdir [pwd]
	    }
	    set file [tk_getOpenFile -filetypes \
		    [list [list [_ "All files"] [list "*"]]] -initialdir \
		    $lastdir \
		    -parent $w -title [_ "Load file"]]
	    if { $file == "" } { return }
	    $lognoter_db addpreference lastdir [file dirname $file]
	    return [fieldtype_file_ops $w $wList open $key $node $file]
	}
	open_files {
	    set file [lindex $args 0 0]
	    return [fieldtype_file_ops $w $wList open $key $node $file]
	}
	paste {
	    set entry [lindex $args 0]
	    set err [catch { clipboard isimage } ret]
	    if { !$err && $ret } {
		set img [clipboard get -type tclimage]
		set data [$img data -format jpeg]
		image delete $img
		fieldtype_file_ops $w $wList open $key $node -data $data \
		    jpeg pastedimage.jpeg
		return -code break
	    }
	}
	ramviewer {
	    set err [catch {
		    set data [cu::inflate [dict get \
		                $uvalues_moredata([$node @n]) contents]]
		    set img [image create photo -data $data]
		} errstring]
	    if { $err } {
		snit_messageBox -message [_ "Could not open RamViewer (%s)" \
		        $errstring] -parent $w
		return
	    }
	    package require RamViewer
	    set w [RamViewer::Init "" $img 1]
	    focus -force [focus -lastfor $w]
	    image delete $img
	}
	open {
	    set file [lindex $args 0]
	    if { $file eq "-data" } {
		foreach "data type file" [lrange $args 1 3] break
		set now [date_tcl2sql [clock seconds]]
		foreach "cdate mdate" [list $now $now] break
		set size [string length $data]
		set contents [cu::deflate -level 9 $data]
		set uvalues($key,[$node @n]) $file
	    } elseif { ![file isdirectory $file] } {
		set fin [open $file r]
		fconfigure $fin -translation binary
		set contents [cu::deflate -level 9 [read $fin]]
		close $fin
		set size [file size $file]
		set uvalues($key,[$node @n]) [file tail $file]
		set type [string tolower [string range [file extension $file] 1 end]]
		file stat $file stat
		set stat_dict [array get stat]
		set cdate [date_tcl2sql [dict get $stat_dict ctime]]
		set mdate [date_tcl2sql [dict get $stat_dict mtime]]
	    } else {
		set zipfile [cu::file::tempfile formulae]
		set pwd [pwd]
		cd [file dirname $file]
		file delete -force $zipfile
		cu::zipfile -r -9 $zipfile [file tail $file]
		cd $pwd
		set fin [open $zipfile r]
		fconfigure $fin -translation binary
		set contents [read $fin]
		close $fin
		set size [file size $zipfile]
		set uvalues($key,[$node @n]) [file tail $file]
		if { ![string equal -nocase [file extension $file] ".gid"] } {
		    set type directory
		} else { set type gid }
		file stat $file stat
		set stat_dict [array get stat]
		set cdate [date_tcl2sql [dict get $stat_dict ctime]]
		set mdate [date_tcl2sql [dict get $stat_dict mtime]]
	    }

	    dict set uvalues_moredata([$node @n]) contents $contents
	    dict set uvalues_moredata([$node @n]) cdate $cdate
	    dict set uvalues_moredata([$node @n]) mdate $mdate
	    dict set uvalues_moredata([$node @n]) type $type
	    dict set uvalues_moredata([$node @n]) size $size
	    dict set uvalues_moredata([$node @n]) path $file

	    fieldtype_file_ops $w $wList update_tooltip $key $node
	}
	clear {
	    set uvalues($key,[$node @n]) ""
	    set uvalues_moredata([$node @n]) ""
	    fieldtype_file_ops $w $wList update_tooltip $key $node
	}
	update_tooltip {
	    set file [string trim $uvalues($key,[$node @n])]
	    if { $file ne "" } {
		set txt [_ "File: %s\n" $file]
	    }
	    if { ![info exists uvalues_moredata([$node @n])] } {
		set uvalues_moredata([$node @n]) ""
	    }
	    set err [catch { dict get $uvalues_moredata([$node @n]) size } size]
	    if { !$err && $size ne "" } {
		set size [dict get $uvalues_moredata([$node @n]) size]
		set mega [expr {1024.0*1024.0}]
		if { $size < $mega } {
		    set sizep [format "%.4g Kb" [expr {$size/1024.0}]]
		} else {
		    set sizep [format "%.4g Mb" [expr {$size/$mega}]]
		}
		append txt [_ "Size: $sizep\n"]
	    }
	    append txt [$node @help ""]
	    foreach widget $wList {
		tooltip $widget $txt
	    }
	}
	execute - extract {
	    set err [catch { dict get $uvalues_moredata([$node @n]) size } size]
	    if { $err || $size eq "" } {
		snit_messageBox -message [_ "There is no file to extract"] \
		    -parent $w
		return
	    }
	    set dir [lindex $args 0]
	    if { $what eq "execute" } {
		set dir [cu::file::tempdir]
	    } elseif { $dir eq "" } {
		set lastdir [$lognoter_db getpreference lastdir]
		set dir [tk_chooseDirectory -initialdir $lastdir \
		        -title [_ "Save file"] \
		        -mustexist 0]
		if { $dir == "" } { return }
		if { ![file exists $dir] } { file mkdir $dir }
		$lognoter_db addpreference lastdir $dir
	    }
	    set file $uvalues($key,[$node @n])
	    set fullfile [file join $dir $file]
	    set type [dict get $uvalues_moredata([$node @n]) type]

	    if { $type eq "directory" || $type eq "gid" } {
		set zipfile [cu::file::tempfile zip]
		set fout [open $zipfile w]
		fconfigure $fout -translation binary
		puts -nonewline $fout [dict get $uvalues_moredata([$node @n]) contents]
		close $fout
		cd $dir
		if { $dir ne [cu::file::tempdir] && [file exists $fullfile] } {
		    set txt [_ "File '%s' exists. Overwrite?" $file]
		    set retval [snit_messageBox -default ok -icon question \
		            -message $txt -parent $w \
		            -type okcancel]
		    if { $retval == "cancel" } { return }
		}
		file delete -force $file
		set pwd [pwd]
		cu::unzipfile -o $zipfile
		cd $pwd
		file delete -force $zipfile
	    } else {
		if { $dir ne [cu::file::tempdir] && [file exists $fullfile] } {
		    set txt [_ "File '%s' exists. Overwrite?" $file]
		    set retval [snit_messageBox -default ok -icon question \
		            -message $txt -parent $w \
		            -type okcancel]
		    if { $retval == "cancel" } { return }
		}
		set fout [open $fullfile w]
		fconfigure $fout -translation binary
		puts -nonewline $fout [cu::inflate [dict get \
		            $uvalues_moredata([$node @n]) contents]]
		close $fout
	    }
	    if { $what eq "execute" } {
		if { $type ne "gid" } {
		    set etype start
		} else {
		    set etype gid
		}
		set err [catch { $lognoter_db executefile $etype $fullfile } errstring]
		if { $err } {
		    snit_messageBox -message [_ "error: %s" $errstring] -parent $w
		    return
		}
	    } else {
		snit_messageBox -message [_ "Saved file in '%s'" $fullfile] -parent $w
	    }
	}
    }
}

proc formulae::fill_form { input_values input_values_more } {
    variable uvalues
    variable uvalues_moredata

    foreach i [array names uvalues_moredata] {
	if { [dict exists $input_values_more $i] } {
	    set uvalues_moredata($i) [dict get $input_values_more $i]
	}
    }
    foreach i [array names uvalues] {
	regexp {,(.*)} $i {} name
	if { [dict exists $input_values $name] } {
	    set uvalues($i) [dict get $input_values $name]
	}
	formulae eval [list set $name $uvalues($i)]
    }
}

proc formulae::window_actualize { args } {
    variable update_stack
    variable svg_stack
    variable uvalues
    variable uvalues_moredata
    variable nicer_name
    variable update_callback

    set optional {
	{ -exec_callback boolean 1 }
	{ -conditions_window wp }
    }
    set compulsory ""
    parse_args $optional $compulsory $args

    foreach i [array names uvalues] {
	regexp {,(.*)} $i {} name
	formulae eval [list set $name $uvalues($i)]
    }

    if { $conditions_window ne "" } {
	set wp $conditions_window
	destroy $wp.conditions_window
	set w [toplevel $wp.conditions_window -bd 2 -relief raised -bg white]
	wm overrideredirect $w 1
	
	wm withdraw $w
	set cmdList [list [list destroy $w]]
	if { [focus] ne "" } {
	    lappend cmdList [list focus [focus]]
	}
	if { [grab current $wp] ne "" } {
	    lappend cmdList [list grab [grab current $wp]]
	}
	bind $w <1> "[join $cmdList ";"]; break"
	bind $w <KeyPress> "[join $cmdList ";"]; break"
	set idx 0
    }

    set conditions_window_results 0

    foreach i $update_stack {
	eval $i

	foreach "cmd ntype name pname value units true false wimg" $i break
	if { $conditions_window ne "" && $cmd eq "calculate_and_update_formula" && 
	    $ntype eq "condition" } {
	    set wimg2 [wordwidgetformula $w.wf[incr idx] -font [$wimg cget -font] \
		    -normalize_spaces 0 -fontsize [$wimg cget -fontsize] \
		    -maxwidth [$wimg cget -maxwidth]]
	    grid $wimg2 -sticky w -padx 2 -pady 2
	    calculate_and_update_formula $ntype $name $pname $value $units $true \
		$false $wimg2
	    incr conditions_window_results
	}
    }
    if { $conditions_window ne "" } {
	if { !$conditions_window_results } {
	    destroy $w
	} else {
	    update idletasks
	    set x [expr {[winfo rootx $wp]+20}]
	    set y [expr {[winfo rooty $wp]+[winfo height $wp]+10}]
	    if { $x+[winfo reqwidth $w] > [winfo screenwidth $w] } {
		set x [expr {[winfo screenwidth $w]-[winfo reqwidth $w]}]
		if { $x < 0 } { set x 0 }
	    }
	    if { $y+[winfo reqheight $w] > [winfo screenheight $w] } {
		set y [expr {[winfo screenheight $w]-[winfo reqheight $w]}]
		if { $y < 0 } { set y 0 }
	    }
	    wm geometry $w +$x+$y
	    wm deiconify $w
	    grab $w
	    focus $w
	}
    }
    set values ""
    foreach i [array names nicer_name] {
	lappend values $i [formulae eval [list set $i]]
    }
    foreach i $svg_stack {
	$i configure -parameters $values
	$i draw
    }
    if { $exec_callback && $update_callback ne "" } {
	eval $update_callback [list $values]
    }
    return [list $values [array get uvalues_moredata]]
}

proc formulae::window_actualize_entry { entry name key isnumber } {
    variable uvalues

    if { $isnumber } {
	set ret [isdouble_or_void $uvalues($key,$name)]
    } else { set ret 1 }
    if { !$ret } {
	bell
	$entry selection range 0 end
	focus $entry
    } else {
	window_actualize -conditions_window $entry
    }
}

################################################################################
#    reports
################################################################################

proc formulae::create_report_from_file { xmlfile } {

    return [create_report [tDOM::xmlReadFile $xmlfile]]
}

# *** not used
proc formulae::report_loadcases_table { root nroot } {

    set setNode [$root selectNodes setname]
    if { $setNode eq "" } { return }

    set xml "<table><row>"
    foreach node [$root selectNodes //param|//condition] {
	append xml "<entry><emphasis role='formula'>[xml [$node @pn [$node @n]]]"
	if { [$node hasAttribute units] } {
	    append xml " ([xml [$node @units]])"
	}
	append xml "</emphasis></entry>"
    }
    append xml </row>

    foreach nameNode [$setNode selectNodes name] {
	if { [$nameNode @print 1] == 0 } { continue }
	foreach nameparamN [$nameNode selectNodes nameparam] {
	    set n [$nameparamN @n]
	    set value [$nameparamN @value]
	    set paramNode [$root selectNodes "//param\[@n='[xml $n]'\]"]
	    $paramNode setAtributte value $value
	}
	append xml "<row>"
	foreach node [$root selectNodes //param|//condition] {
	    if { [string index [$node @value] 0] == "\[" } {
		set value [string range [$node @value] 1 end-1]
		set v [eval $value]
	    } else {
		set v [expr {[$node @value]*1.0}]
	    }
	    if { [$node nodeName] eq "param" } {
		formulae eval [list set [$node @n] $v]
		append xml "<entry>[format %.4g $v]</entry>"
	    } else {
		append xml "<entry><emphasis role='formula'>"
		if { $v } {
		    if { [$node @true ""] ne "" } {
		        append xml [xml "[$node @true] (ok)"]
		    } else {
		        append xml [xml "[_ OK#C#Valid] (ok)"]
		    }
		} else {
		    if { [$node @false ""] ne "" } {
		        append xml [xml "[$node @false] (fail)"]
		    } else {
		        append xml [xml "[_ {Fail#C#No Valid}] (fail)"]
		    }
		}
		append xml "</emphasis></entry>"
	    }
	}
	append xml "</row>"
    }
    append xml "</table>"
    
    $nroot appendXML $xml
}

proc formulae::report_loadcases { root nroot } {

    set setNode [$root selectNodes setname]
    if { $setNode eq "" } { return }

    foreach nameNode [$setNode selectNodes name] {
	if { [$nameNode @print 1] == 0 } { continue }
	if { [$setNode @n] eq [$nameNode @n] } { continue }
	foreach nameparamN [$nameNode selectNodes name_param] {
	    set n [$nameparamN @n]
	    set value [$nameparamN @value]
	    set paramNode [$root selectNodes "//param\[@n='[xml $n]'\]"]
	    $paramNode setAttribute value $value
	}
	set em [$nroot appendChildTag emphasis]
	$em setAttribute role header
	$em appendChildText [_ "Load case %s" [$nameNode @n]]
	$nroot appendChildTag para
	foreach node [$root selectNodes container] {
	    switch [$node nodeName] {
		container {
		    if { [$node @condition ""] ne "" } {
		        set ret [eval [string range [$node @condition] 1 end-1]]
		        if { !$ret } { continue }
		    }
		    set ns { svg http://www.w3.org/2000/svg }
		    set svgNode [$node selectNodes -namespaces $ns svg:svg]
		    if { $svgNode ne "" } {
		        set m [$nroot appendChildTag mediaobject imageobject]
		        $m appendXML [$svgNode asXML]
		        set svgNode [$m selectNodes -namespaces $ns svg:svg]
		        $svgNode setAttribute width 50% height 50%
		    }
		    set em [$nroot appendChildTag emphasis]
		    $em setAttribute role header
		    $em appendChildText [$node @pn [$node @n]]
		    set tag [$nroot appendChildTag table]
		    $tag setAttribute \
		        relativesizes "0 65 35" tabstyle "" \
		        cols_justify "left left"
		    process_container $tag $node
		}
	    }
	}
    }
}

proc formulae::create_report { xmldata } {
    variable page_tcl_code
    variable lognoter_db
    variable window_doc

    dom parse $xmldata doc

    if { [info exists window_doc] } {
	set window_doc_save $window_doc
    }
    set window_doc $doc

    if { [interp exists formulae] } { interp delete formulae }
    interp create formulae

    formulae eval [list set auto_path $::auto_path]
    #interp alias formulae ::tcl::mathfunc::defaultvalue "" formulae::defaultvalue $root
    interp alias formulae ::tcl::mathfunc::max "" formulae::mathfunc_max
    interp alias formulae ::tcl::mathfunc::min "" formulae::mathfunc_min
    interp alias formulae ::tcl::mathfunc::case "" formulae::mathfunc_case
    interp alias formulae ::tcl::mathfunc::pi "" formulae::mathfunc_pi
    interp alias formulae ::tcl::mathfunc::interpolate "" formulae::mathfunc_interpolate
    interp alias formulae ::tcl::mathfunc::dictionary "" formulae::mathfunc_dictionary
    interp alias formulae xml "" formulae::xml
    if { $lognoter_db ne "" } {
	interp alias formulae lognoter_db "" $lognoter_db
	interp alias formulae executefile "" $lognoter_db executefile
	interp alias formulae sql "" $lognoter_db sql
	formulae eval [list proc refresh {} {}]
    }

    catch {
	formulae eval $page_tcl_code
    }

    set ndoc [dom createDocument report]
    $doc documentElement root
    $ndoc documentElement nroot
    $nroot setAttribute version 1.0

    set printedLC_title 0
    foreach node [$root selectNodes title|description|container] {
	switch [$node nodeName] {
	    title {
		set em [$nroot appendChildTag emphasis]
		$em setAttribute role header
		$em appendChildText [$node text]
	    }
	    description {
		set para [$nroot appendChildTag para]
		$para appendChildText [$node text]

	    }
	    container {
		set xp {../setname/@n = ../setname/name[@print='0']/@n}
		if { [$node selectNodes $xp] } { continue }

		if { !$printedLC_title } {
		    set setNode [$root selectNodes setname]
		    if { $setNode ne "" && [$setNode @n ""] ne "" } {
		        set em [$nroot appendChildTag emphasis]
		        $em setAttribute role header
		        $em appendChildText [_ "Load case %s" [$setNode @n]]
		        $nroot appendChildTag para
		    }
		    set printedLC_title 1
		}
		if { [$node @condition ""] ne "" } {
		    set ret [eval [string range [$node @condition] 1 end-1]]
		    if { !$ret } { continue }
		}
		set ns { svg http://www.w3.org/2000/svg }
		set svgNode [$node selectNodes -namespaces $ns svg:svg]
		if { $svgNode ne "" } {
		    set m [$nroot appendChildTag mediaobject imageobject]
		    $m appendXML [$svgNode asXML]
		    set svgNode [$m selectNodes -namespaces $ns svg:svg]
		    $svgNode setAttribute width 50% height 50%
		}
		set em [$nroot appendChildTag emphasis]
		$em setAttribute role header
		$em appendChildText [$node @pn [$node @n]]
		set tab [$nroot appendChildTag table]
		$tab setAttribute \
		    relativesizes "0 65 35" tabstyle "" \
		    cols_justify "left left"
		process_container $tab $node
	    }
	}
    }
    report_loadcases $root $nroot

    if { [info exists window_doc_save] } {
	set window_doc $window_doc_save
    } else {
	unset window_doc
    }
    return $ndoc
}

proc formulae::create_report_xml { xmldata } {
    set ndoc [create_report $xmldata]
    set _ ""
    set root [$ndoc documentElement]
    foreach node [$root childNodes] {
	append _ [$node asXML]
    }
    return $_
}

proc formulae::print_report { ndoc file format } {

    package require xml2pdf

    switch $format {
	pdf - print {
	    PDFWriter::choosetype $format
	    
	    set papersize A4
	    set boxList {
		{mainbox {225 285 1505 2260} {11.5% 10.4% 89.3% 92.9%} 1}
		{colorbox all {10.4% 9.47% 90.5% 93.35%} {} #CEC2C2 3}
		{colorbox all {10.2% 9.62% 90.3% 93.5%} {} #3A967E 3}
		{text all {11% 7.3% 65% 8.5%} Helvetica 12 black white sw {<wordwidget><para><emphasis role="strong">$sectiontitle</emphasis></para></wordwidget>} 1}
		{text all {10% 94.1% 25% 95.8%} Helvetica 10 #000000 white nw {<wordwidget><para>Pgina $pagenum</para></wordwidget>} 1}
		{text all {66.6% 94.1% 83.6% 96%} Helvetica 6 #000000 white n {<wordwidget><para>Developed by Compass, Ing. y Sist.</para><para>http://www.compassis.com</para></wordwidget>} 0}
	    }
	    set opts {
		TitleBackgroundColor #dceae6
		sizedefault 10
	    }
	    if { $file == "" } {
		set file [xml2pdf::PrintXML [$ndoc asXML -indent none] $papersize \
		        $boxList $opts]
	    } else {
		lappend opts filename [file root $file].pdf
		xml2pdf::PrintXML [$ndoc asXML -indent none] $papersize $boxList $opts
	    }
	}
	html {
	    set opts [list filename $file multiplesfiles $multiplesfiles]
	    #lappend opts outputrtf 1
	    xml2html:::PrintHTML [$ndoc asXML -indent none] $opts
	}
	rtf {
	    set opts [list filename $file]
	    lappend opts outputrtf 1
	    xml2html:::PrintHTML [$ndoc asXML -indent none] $opts
	}
	xml {
	    if { $file eq "" } {
		foreach "file fout" [xml2html::GiveTempFile htmfile .html] break
	    } else {
		set fout [open $file w]
	    }
	    fconfigure $fout -encoding utf-8
	    puts $fout "<?xml version=\"1.0\" encoding=\"utf-8\"?><!-- -*- coding: utf-8;-*- -->"
	    set root [$ndoc documentElement]
	    puts $fout "<!DOCTYPE [$root nodeName]>"
	    $ndoc asXML -indent none -channel $fout
	    close $fout
	}
    }
}

################################################################################
#    calc_and_update_tree
################################################################################

proc formulae::add_row_to_database { args } {

    set optional {
    }
    set compulsory "db table values values_moredata root"
    parse_args $optional $compulsory $args

    set cols [$db sql column_names $table]
    set cols_r [lrange $cols 3 end]
    set valuesList [lrepeat [expr {[llength $cols]-3}] null]

    foreach "n v" $values {
	set ipos [lsearch -exact $cols_r $n]
	if { $ipos == -1 } { continue }
	set v [string trim $v]
	
	set xp [format_xpath {string(//param[@n=%s]/@field_type)} $n]
	set field_type [$root selectNodes $xp]
	if { $field_type eq "file" } {
	    set err [catch { dict get $values_moredata $n contents } c]
	    if { $err || $c eq "" } {
		set v ""
	    } elseif { $v eq "" } {
		dict set values_moredata $n ""
	    }
	}
	if { [string is double -strict $v] } {
	    lset valuesList $ipos [$db sql escape $v]
	} else {
	    lset valuesList $ipos '[$db sql escape $v]'
	}
    }
    foreach "n v" $values_moredata {
	dict for "n2 v2" $v {
	    set nfull "${n}__$n2"
	    set ipos [lsearch -exact $cols_r $nfull]
	    if { $ipos == -1 } { continue }
	    lset valuesList $ipos '[$db sql escape $v2]'
	}
    }


    $db sql check_state 

    set cmd "insert into \"$table\" (\"[join $cols \",\"]\") values(
	null,now(),now(),"
    append cmd [join $valuesList ,] ")"
    $db sql exec $cmd
}

proc formulae::calc_and_update_tree { args } {

    set optional {
	{ -lognoter_db db "" }
    }
    set compulsory "doc values_dict"
    parse_args $optional $compulsory $args

    set page_tcl_code ""
    foreach node [$doc selectNodes //tcl] {
	append page_tcl_code "[$node text]\n"
    }

    if { [interp exists formulae] } { interp delete formulae }
    interp create formulae

    formulae eval [list set auto_path $::auto_path]
    #interp alias formulae ::tcl::mathfunc::defaultvalue "" formulae::defaultvalue $root
    interp alias formulae ::tcl::mathfunc::max "" formulae::mathfunc_max
    interp alias formulae ::tcl::mathfunc::min "" formulae::mathfunc_min
    interp alias formulae ::tcl::mathfunc::case "" formulae::mathfunc_case
    interp alias formulae ::tcl::mathfunc::pi "" formulae::mathfunc_pi
    interp alias formulae ::tcl::mathfunc::interpolate "" formulae::mathfunc_interpolate
    interp alias formulae ::tcl::mathfunc::dictionary "" formulae::mathfunc_dictionary
    interp alias formulae xml "" formulae::xml
    if { $lognoter_db ne "" } {
	interp alias formulae lognoter_db "" $lognoter_db
	interp alias formulae executefile "" $lognoter_db executefile
	interp alias formulae sql "" $lognoter_db sql
	formulae eval [list proc refresh {} {}]
    }
    catch {
	formulae eval $page_tcl_code
    }
    foreach node [$doc selectNodes //param|//condition] {
	if { [string index [$node @value] 0] == "\[" } {
	    set value [string range [$node @value] 1 end-1]
	    set valueE [string map [list \[ \\\[] $value]

	    if { [lsearch [list "formatted expression" "text expression"] \
		[$node @field_type ""]] != -1 } {
		set err [catch { formulae eval $valueE } v]
	    } else {
		set err [catch { eval $valueE } v]
	    }
	    if { $err } {
		set v "!ERROR ($v)"
	    }
	} else {
	    if { [dict exists $values_dict [$node @n]] } {
		set v [dict get $values_dict [$node @n]]
	    } else {
		set v [$node @value]
	    }
	    set err [catch { expr {$v*1.0} } newvalue]
	    if { !$err } { set v $newvalue }
	}
	dict set values_dict [$node @n] $v

	if { [$node nodeName] eq "param" } {
	    formulae eval [list set [$node @n] $v]
	    if { [string match "formatted*" [$node @field_type ""]] } {
		if { [string trim $v] eq "" } { set v "<lognoter/>" }
		$node appendXML $v
		set n0 [lindex [$node childNodes] 0]
		foreach n [$n0 childNodes] {
		    $node appendChild $n
		}
		$n0 delete
		$node setAttribute value ""
	    } else {
		$node setAttribute value $v
	    }
	} else {
	    #nothing by now
	}
    }

    set root [$doc selectNodes //formulae]

    if { [$root @database ""] ne "" } {
	add_row_to_database $lognoter_db [$root @database] $values_dict \
	    "" $root
    }
}

if { 0 } {
    set ndoc [formulae::create_report_from_file ~/../Escritorio/placa_anclaje.xml]
    formulae::print_report $ndoc placa_anclaje.pdf pdf
    exit
}


















